import argparse
from nanomock.nanomock_manager import NanoLocalManager
from nanomock.internal.utils import is_packaged_version
from pathlib import Path
from os import environ
import json


def _get_default_app_dir():
    if environ.get("NL_CONF_DIR"):
        return Path(environ.get("NL_CONF_DIR"))
    if is_packaged_version():
        return Path.cwd()
    else:
        return Path.cwd() / 'nanomock'


def parse_args():
    parser = argparse.ArgumentParser(description='nanomock cli')
    parser.add_argument('command',
                        choices=[
                            'create', 'start', 'start_nodes', 'status',
                            'restart', 'init', 'init_wallets', 'conf_edit',
                            'stop', 'stop_nodes', 'update', 'remove', 'reset',
                            'down', 'destroy', 'rpc'
                        ])
    parser.add_argument('--path',
                        default=_get_default_app_dir(),
                        help='Path to the working directory')
    parser.add_argument(
        '--nodes',
        nargs='+',
        help=
        'List of container names (only required for start, stop, remove, and rpc commands)'
    )
    parser.add_argument(
        '--project_name',
        default="nanomock",
        help='project_name for docker-compose to know what to execute')

    # Add the optional --payload argument
    parser.add_argument(
        '--payload',
        type=json.loads,
        help="JSON request payload (only required for rpc command)")

    return parser.parse_args()


def main(args=None):
    args = args or parse_args()
    manager = NanoLocalManager(args.path, args.project_name,
                               environ.get("NL_CONF_FILE", "nl_config.toml"))

    manager.execute_command(args.command, args.nodes, args.payload)


if __name__ == '__main__':
    main()



from .interface import DockerInterface
from .linux import LinuxDockerInterface
from .windows import WindowsDockerInterface
from .macos import MacDockerInterface
from typing import Type
import platform


def create_docker_interface(compose_path: str,
                            project_name: str) -> DockerInterface:
    os_name = platform.system()

    if os_name == "Windows":
        return WindowsDockerInterface(compose_path, project_name)
    elif os_name == "Linux":
        return LinuxDockerInterface(compose_path, project_name)
    elif os_name == "Darwin":
        return MacDockerInterface(compose_path, project_name)
    else:
        raise NotImplementedError("Unsupported OS")


def get_docker_interface_class() -> Type[DockerInterface]:
    os_name = platform.system()

    if os_name == "Windows":
        return WindowsDockerInterface
    elif os_name == "Linux":
        return LinuxDockerInterface
    elif os_name == "Darwin":
        return MacDockerInterface
    else:
        raise NotImplementedError("Unsupported OS")


from .interface import DockerInterface
from .mixin import DockerMixin
import subprocess


class WindowsDockerInterface(DockerMixin, DockerInterface):

    @classmethod
    def get_docker_gateway_ip(cls) -> str:
        command = "ipconfig | findstr /C:'Docker' /A:1 | findstr /C:'IPv4'"
        output = subprocess.check_output(command, shell=True, text=True)
        return output.strip().split()[-1]


import os
import glob
from pathlib import Path
from typing import List, Optional
from nanomock.internal.utils import get_mock_logger

logger = get_mock_logger()


class OSOperationsMixin:

    def delete_nodes_data(self, path, nodes: Optional[List[str]] = None):

        nodes_to_process = nodes or ['.']
        for node in nodes_to_process:
            node_path = Path(path) / node if nodes else Path(path)

            for filename in glob.glob(str(node_path / '**' / '*.ldb'),
                                      recursive=True):
                try:
                    os.remove(filename)
                except OSError as err:
                    logger.error("Error removing file %s\n%s", filename, err)

    @staticmethod
    def makedirs(path):
        os.makedirs(path, exist_ok=True)



import time
import unittest
from itertools import islice
from math import ceil, floor

from nanomock.modules.nl_rpc import NanoRpc
from nanomock.modules.nl_parse_config import ConfigParser, ConfigReadWrite
from nanomock.modules.nl_nanolib import NanoLibTools, raw_high_precision_multiply


class BlockGenerator():

    single_change_rep = None

    def __init__(self,
                 rpc_url,
                 broadcast_blocks=False,
                 rpc_user=None,
                 rpc_password=None,
                 log_to_console=False):

        self.conf_p = ConfigParser()
        rpc_url = self.conf_p.get_nodes_rpc(
        )[0] if rpc_url is None else rpc_url

        self.broadcast = broadcast_blocks
        self.log_to_console = log_to_console
        self.single_account_open_counter = 0
        self.nano_lib = NanoLibTools()
        self.nano_rpc_default = NanoRpc(rpc_url,
                                        username=rpc_user,
                                        password=rpc_password)

    def get_nano_rpc_default(self):
        return self.nano_rpc_default

    def blockgen_single_account_opener(
            self,
            representative=None,
            source_key=None,  #
            source_seed=None,
            source_index=None,
            destination_key=None,  #
            destination_seed=None,
            destination_index=None,
            send_amount=1,
            number_of_accounts=1000,
            nano_rpc=None,
            accounts_keep_track=False,
            increment_index=False):
        nano_rpc = self.get_nano_rpc_default()
        if accounts_keep_track:
            if self.single_account_open_counter >= number_of_accounts:
                return []
            if increment_index:
                destination_index = self.single_account_open_counter
        self.single_account_open_counter = self.single_account_open_counter + 1

        destination = self.nano_lib.nanolib_account_data(
            private_key=destination_key,
            seed=destination_seed,
            index=destination_index)
        source = self.nano_lib.nanolib_account_data(private_key=source_key,
                                                    seed=source_seed,
                                                    index=source_index)

        send_block = nano_rpc.create_send_block_pkey(source["private"],
                                                     destination["account"],
                                                     send_amount,
                                                     broadcast=self.broadcast)

        open_block = nano_rpc.create_open_block(destination["account"],
                                                destination["private"],
                                                send_amount,
                                                representative,
                                                send_block["hash"],
                                                broadcast=self.broadcast)
        open_block["account_data"]["source_seed"] = destination_seed
        open_block["account_data"]["source_index"] = destination_index

        res = [send_block, open_block]
        if self.log_to_console:
            print("accounts opened:  {:>6}".format(
                self.single_account_open_counter),
                  end='\r')
        return res

    def set_single_change_rep(self, rep=None, nano_rpc: NanoRpc = None):
        #returns random rep if rep is not specified
        if rep is not None: self.single_change_rep = rep
        elif rep is None and nano_rpc is not None:
            self.single_change_rep = nano_rpc.get_account_data(
                nano_rpc.generate_seed(), 0)["account"]
        else:
            nano_rpc = self.get_nano_rpc_default()
            self.single_change_rep = nano_rpc.get_account_data(
                nano_rpc.generate_seed(), 0)["account"]
        return self.single_change_rep

    def blockgen_single_change(self,
                               source_seed=None,
                               source_index=None,
                               source_private_key=None,
                               rep=None,
                               nano_rpc=None):
        nano_rpc = self.get_nano_rpc_default()
        if rep is None: rep = self.single_change_rep
        if rep is None:
            rep = nano_rpc.get_account_data(nano_rpc.generate_seed(),
                                            0)["account"]

        if source_private_key is not None:
            return nano_rpc.create_change_block_pkey(source_private_key,
                                                     rep,
                                                     broadcast=self.broadcast)
        elif source_seed is not None and source_index is not None:
            return nano_rpc.create_change_block(source_seed,
                                                source_index,
                                                rep,
                                                broadcast=self.broadcast)
        else:
            raise ValueError(
                f"Either source_private_key({source_private_key})   OR   source_seed({source_seed}) and source_index({source_index}) must not be None"
            )

    def recursive_split(self, source_account_data, destination_seed,
                        representative, number_of_accounts, splitting_depth,
                        current_depth, final_account_balance_raw, split_count):

        blocks_current_depth = self.blockgen_single_account_opener(
            representative=representative,
            source_key=source_account_data["private"],
            destination_seed=destination_seed,
            #destination_index=source_dest_account_data["index"] + 1,
            accounts_keep_track=True,
            increment_index=True,
            number_of_accounts=number_of_accounts,
            send_amount=int(
                raw_high_precision_multiply(
                    (split_count**(splitting_depth - current_depth + 1) -
                     split_count) + 1, final_account_balance_raw)))

        if len(blocks_current_depth) == 0:
            return blocks_current_depth

        blocks_next_depth = self.blockgen_account_splitter(
            number_of_accounts=number_of_accounts,
            destination_seed=destination_seed,
            current_depth=current_depth + 1,
            representative=representative,
            source_private_key=blocks_current_depth[1]["account_data"]
            ["private"],
            final_account_balance_raw=final_account_balance_raw,
            split_count=split_count)
        return blocks_current_depth + blocks_next_depth  #blocks_current_depth.extends(blocks_next_depth)

    def get_spliting_depth(self, number_of_accounts, split_count):
        sum_l = 0
        for exponent in range(1, 128):
            sum_l = sum_l + (split_count**exponent)
            if sum_l >= number_of_accounts: break
        return exponent

    def get_accounts_for_depth(self, split_count, splitting_depth):
        accounts = 0
        for i in range(1, splitting_depth + 1):
            accounts = accounts + (split_count**i)
        return accounts

    def blockgen_account_splitter(self,
                                  source_private_key=None,
                                  source_seed=None,
                                  source_index=0,
                                  destination_seed=None,
                                  number_of_accounts=1000,
                                  current_depth=1,
                                  split_count=2,
                                  representative=None,
                                  final_account_balance_raw=10**30,
                                  nano_rpc=None):
        '''create {split_count} new accounts from 1 account recursively until {number_of_accounts} is reached.
           each account sends its funds to {split_count}  other accounts and keeps a minimum balance of {final_account_balance_raw}
           returns 2 * {number_of_accounts} blocks
           '''

        splitting_depth = self.get_spliting_depth(number_of_accounts,
                                                  split_count)

        if current_depth > splitting_depth:
            return []  #end of recursion is reached
        nano_rpc = self.get_nano_rpc_default()

        source_account_data = self.nano_lib.nanolib_account_data(
            private_key=source_private_key,
            seed=source_seed,
            index=source_index)
        if current_depth == 1:
            max_accounts_for_depth = self.get_accounts_for_depth(
                split_count, splitting_depth)
            print(
                f"Creating {number_of_accounts} of {max_accounts_for_depth} possible accounts for current splitting_depth : {splitting_depth} and split_count {split_count}"
            )
            self.single_account_open_counter = 0  #reset variable when multiple tests run successively
            unittest.TestCase().assertGreater(
                int(
                    nano_rpc.check_balance(
                        source_account_data["account"])["balance_raw"]),
                int(
                    raw_high_precision_multiply(number_of_accounts,
                                                final_account_balance_raw)))
            if representative is None:  #keep the same representative for all blocks
                representative = nano_rpc.account_info(
                    source_account_data["account"]
                )["representative"]  #keep the same representative for all opened accounts

        all_blocks = []
        for _ in range(0, split_count):
            all_blocks.extend(
                self.recursive_split(source_account_data, destination_seed,
                                     representative, number_of_accounts,
                                     splitting_depth, current_depth,
                                     final_account_balance_raw, split_count))

        if current_depth == 1:
            self.single_account_open_counter = 0  #reset counter for next call
        return all_blocks

    def get_hashes_from_blocks(self, blocks):
        if isinstance(blocks, list):
            block_hashes = [x["hash"] for x in blocks]
            return block_hashes
        elif isinstance(blocks, dict):
            return blocks.get("hash", "")

    def make_deep_forks(self,
                        source_seed,
                        source_index,
                        dest_seed,
                        amount_raw,
                        peer_count,
                        forks_per_peer=1,
                        max_depth=5,
                        current_depth=0):
        fork_blocks = {"gap": [], "forks": []}
        nano_rpc = self.get_nano_rpc_default()
        send_block = nano_rpc.create_block(
            "send",
            source_seed=source_seed,
            source_index=source_index,
            destination_account=nano_rpc.get_account_data(
                source_seed, source_index)["account"],
            amount_raw=amount_raw,
            read_in_memory=False,
            add_in_memory=True)

        fork_blocks["gap"].append(nano_rpc.get_block_result(send_block, False))

        fork_blocks["forks"] = self.recursive_fork_depth(
            source_seed,
            source_index,
            dest_seed,
            amount_raw,
            peer_count,
            forks_per_peer=forks_per_peer,
            max_depth=max_depth,
            current_depth=current_depth)

        return fork_blocks

    def recursive_fork_depth(self,
                             source_seed,
                             source_index,
                             dest_seed,
                             amount_raw,
                             peer_count,
                             forks_per_peer=1,
                             max_depth=5,
                             current_depth=0):
        res = []
        if current_depth >= max_depth: return res

        nano_rpc = self.get_nano_rpc_default()
        next_depth = current_depth + 1
        current_dest_start_index = (forks_per_peer * peer_count *
                                    next_depth) + next_depth
        previous_dest_start_index = (forks_per_peer * peer_count *
                                     current_depth) + current_depth

        #current_dest_start_index = (100 ** next_depth)
        #previous_dest_start_index = 100 ** current_depth

        for i in range(0, forks_per_peer * peer_count):
            dest_index = current_dest_start_index + i
            dest_account = nano_rpc.get_account_data(dest_seed,
                                                     dest_index)["account"]
            send_block = nano_rpc.create_block(
                "send",
                source_seed=source_seed,
                #source_index=source_index if current_depth == 0 else 100 ** current_depth,
                source_index=source_index
                if current_depth == 0 else previous_dest_start_index,
                destination_account=dest_account,
                amount_raw=amount_raw,
                read_in_memory=True,
                add_in_memory=False)

            receive_block = nano_rpc.create_block(
                "receive",
                source_seed=dest_seed,
                source_index=dest_index,
                destination_account=dest_account,
                representative=dest_account,
                amount_raw=amount_raw,
                link=send_block["hash"],
                read_in_memory=False,
                add_in_memory=True,
            )
            res.append(nano_rpc.get_block_result(send_block, False))
            res.append(nano_rpc.get_block_result(receive_block, False))

            next_res = self.recursive_fork_depth(dest_seed,
                                                 dest_index,
                                                 dest_seed,
                                                 amount_raw,
                                                 peer_count,
                                                 max_depth=max_depth,
                                                 current_depth=next_depth)
            res.extend(next_res)
        # print(
        #     ">>>>DEBUG", "Current_depth : {:>2}  results {:>6}".format(
        #         current_depth, len(res)))
        return res


class BlockAsserts():
    from multiprocessing import Value

    tc = unittest.TestCase()

    def __init__(self, rpc_url=None):
        self.conf_p = ConfigParser()
        rpc_url = self.conf_p.get_nodes_rpc(
        )[0] if rpc_url is None else rpc_url
        self.nano_rpc_default = NanoRpc(rpc_url)
        self.nano_rpc_all = self.get_rpc_all()

    def get_rpc_all(self):
        return [NanoRpc(x) for x in self.conf_p.get_nodes_rpc()]

    def get_rpc_custom(self, rpc_urls):
        if rpc_urls is None: return self.get_rpc_all()
        return [NanoRpc(x) for x in rpc_urls]

    def assert_list_of_blocks_published(self,
                                        list_of_blocks,
                                        sync=True,
                                        is_running=Value('i', False),
                                        tps=450):
        for blocks in list_of_blocks:
            self.assert_blocks_published(blocks, sync=sync, tps=tps)
        is_running.value = False

    def assert_blocks_published(self, blocks, sync=True, tps=450):
        blocks_to_publish_count = len(blocks)
        rpc_block_count_start = int(
            self.nano_rpc_default.block_count()["count"])

        start_time = time.time()
        interval = 1

        for i in range(0, ceil(blocks_to_publish_count / tps)):
            blocks_subset = list(islice(blocks, i * tps, i * tps + tps))
            self.nano_rpc_default.publish_blocks(
                blocks_subset, json_data=True,
                sync=sync)  #we don't care about the result
            sleep_duration = max(
                0, start_time + (i + 1) * interval - time.time() +
                ((len(blocks_subset) - tps) / tps))
            time.sleep(sleep_duration)

        self.assert_expected_block_count(blocks_to_publish_count +
                                         rpc_block_count_start)

    def assert_expected_block_count(self, expected_count, exit_after_s=5):
        start_time = time.time()
        #with timeout(exit_after_s, exception=RuntimeError) :
        while True:
            if time.time() - start_time > exit_after_s: break
            rpc_block_count_end = int(
                self.nano_rpc_default.block_count()["count"])
            if rpc_block_count_end >= expected_count: break
            time.sleep(0.2)
        self.tc.assertGreaterEqual(
            rpc_block_count_end,
            expected_count)  #if other blocks arrive in the meantime

    def assert_increasing_block_count(self, expected_count, exit_after_s=5):
        stall_time = time.time()
        #with timeout(exit_after_s, exception=RuntimeError) :
        prev_block_count = self.nano_rpc_default.block_count()["count"]
        while True:
            if time.time() - stall_time > exit_after_s: break
            rpc_block_count_end = int(
                self.nano_rpc_default.block_count()["count"])
            if rpc_block_count_end > prev_block_count:  #reset stall_time on increased block_count
                stall_time = time.time()
            if rpc_block_count_end >= expected_count: break
            time.sleep(0.2)

        self.tc.assertGreaterEqual(
            rpc_block_count_end,
            expected_count)  #if other blocks arrive in the meantime

    def assert_single_block_confirmed(self,
                                      hash,
                                      sleep_on_stall_s=0.1,
                                      exit_after_s=120,
                                      exit_on_first_stall=False):
        #Convert hash_string into list of 1 hash and reuse existing method that handles lists
        block_hashes = []
        block_hashes.append(hash)
        return self.assert_blocks_confirmed(
            block_hashes,
            sleep_on_stall_s=sleep_on_stall_s,
            exit_after_s=exit_after_s,
            exit_on_first_stall=exit_on_first_stall)

    def assert_blocks_confirmed(self,
                                block_hashes,
                                max_stall_duration_s=6 * 60,
                                sleep_on_stall_s=5,
                                stall_timeout_max=30 * 60,
                                exit_after_s=60 * 60,
                                exit_on_first_stall=False,
                                log_to_console=False):
        start_time = time.time()
        block_count = len(block_hashes)
        print(">>>DEBUG assert_blocks_confirmed:", block_hashes[0],
              block_count)

        timeout_inc = 0
        try:
            #with timeout(exit_after_s, exception=RuntimeError) :
            confirmed_count = 0
            while confirmed_count < block_count:
                if time.time() - start_time > exit_after_s: break
                last_confirmed_count = confirmed_count
                confirmed_hashes = self.nano_rpc_default.block_confirmed_aio(
                    block_hashes,
                    ignore_errors=["Block not found"],
                )
                block_hashes = list(set(block_hashes) - confirmed_hashes)
                confirmed_count = confirmed_count + len(confirmed_hashes)
                if confirmed_count != block_count:
                    if log_to_console:
                        print(
                            f"{confirmed_count}/{block_count} blocks confirmed",
                            end="\r")
                    time.sleep(sleep_on_stall_s)
                if confirmed_count == last_confirmed_count:  # stalling block_count
                    if exit_on_first_stall:
                        return {
                            "total_block_count": block_count,
                            "confirmed_count": confirmed_count,
                            "unconfirmed_count": block_count - confirmed_count
                        }

                    stall_timeout_max = stall_timeout_max - sleep_on_stall_s
                    stall_timeout_max = timeout_inc + sleep_on_stall_s
                    if timeout_inc >= max_stall_duration_s:
                        raise ValueError(
                            f"No new confirmations for {max_stall_duration_s}s... Fail blocks_confirmed"
                        )  #break if no new confirmatiosn for 6 minutes (default)
                else:  #reset stall timer
                    timeout_inc = 0
                if stall_timeout_max <= 0:
                    raise ValueError(
                        f"Max timeout of {stall_timeout_max} seconds reached")
            first_hash = list(confirmed_hashes
                              )[0] if confirmed_count > 0 else block_hashes[0]
            print(
                f"{confirmed_count}/{block_count} blocks confirmed in {round(time.time() - start_time, 2)} s [{first_hash}] ... [{self.nano_rpc_default.RPC_URL}]"
            )
        except RuntimeError as re:  #when timeout hits
            self.tc.fail(str(re))
        except ValueError as ve:
            self.tc.fail(str(ve))

        self.tc.assertEqual(confirmed_count, block_count)
        return confirmed_count

    def assert_expected_cemented(
        self,
        expected_count,
        rpc_urls=None,
        exit_after_s=30,
    ):
        rpcs = self.get_rpc_custom(rpc_urls)
        break_on_ext_iteration = False
        start_time = time.time()

        while True:
            if break_on_ext_iteration: break
            min_cemented = 1e38
            max_count = 1
            time.sleep(0.2)
            for nano_rpc in rpcs:
                block_count = nano_rpc.block_count()
                min_cemented = min(min_cemented, int(block_count["cemented"]))
                max_count = max(max_count, int(block_count["count"]))
            if expected_count == min_cemented:
                #prints the report when 100% is reached
                break_on_ext_iteration = True
            if time.time() - start_time > exit_after_s:
                self.tc.fail(
                    f"Timout after {exit_after_s}s {min_cemented}/{expected_count} blocks confirmed "
                )
            print("cemented {:>8}/{:<8}  expected: {} [{:4}%]".format(
                min_cemented,
                max_count,
                expected_count,
                floor(min_cemented / max_count * 10000) / 100,
                end="\r"))
            time.sleep(2)

    def assert_all_blocks_cemented(self):
        for nano_rpc in self.nano_rpc_all:
            block_count = nano_rpc.block_count()
            self.tc.assertEqual(block_count["count"], block_count["cemented"])
        return block_count

    @staticmethod
    def assert_blockgen_succeeded(blocks):
        tc = unittest.TestCase()
        if isinstance(blocks, list):
            tc.assertEqual(len(list(filter(lambda x: x["success"], blocks))),
                           len(blocks))
        elif isinstance(blocks, dict):
            tc.assertTrue(blocks["success"])
        else:
            tc.fail("Blocks must be of list or dict type")

    def network_status(self, nodes_name: list = None):
        if nodes_name == []: return ""

        max_count = 0
        nodes_block_count = []

        if nodes_name is not None:
            nodes_rpc = [
                NanoRpc(self.conf_p.get_node_rpc(node)) for node in nodes_name
            ]
        else:
            nodes_rpc = self.nano_rpc_all

        for nano_rpc in nodes_rpc:
            block_count = nano_rpc.block_count()
            version_rpc_call = nano_rpc.version()
            max_count = int(block_count["count"]) if int(
                block_count["count"]) > max_count else max_count
            node_name = self.conf_p.get_node_name_from_rpc_url(nano_rpc)
            nodes_block_count.append({
                "node_name": node_name,
                "count": block_count["count"],
                "cemented": block_count["cemented"]
            })

        node_version = f'{version_rpc_call["node_vendor"]} {version_rpc_call["build_info"].split(" ")[0]}'[
            0:20]

        report = [
            '{:<16} {:<20} {:>6.2f}% synced | {}/{} blocks cemented'.format(
                bc["node_name"], node_version,
                floor(int(bc["cemented"]) / max_count * 10000) / 100,
                bc["cemented"], bc["count"]) for bc in nodes_block_count
        ]
        return '\n' + '\n'.join(report)


class BlockReadWrite():

    def __init__(self):
        self.conf_rw = ConfigReadWrite()

    def read_blocks_from_disk(self,
                              path,
                              seeds=False,
                              hashes=False,
                              blocks=False):
        res = self.conf_rw.read_json(path)
        if seeds: return res["s"]
        if hashes: return res["h"]
        if blocks: return res["b"]
        return res

    def write_blocks_to_disk(self, rpc_block_list, path):
        hash_list = []
        seed_list = []
        block_list = []

        if self.is_nested_list(rpc_block_list):
            for block_list_i in rpc_block_list:
                result = self.process_rpc_block_list(block_list_i)
                hash_list.append(result["hash_list"])
                seed_list.append(result["seed_list"])
                block_list.append(result["block_list"])
        else:
            result = self.process_rpc_block_list(rpc_block_list)
            hash_list.append(result["hash_list"])
            seed_list.append(result["seed_list"])
            block_list.append(result["block_list"])

        res = {"h": hash_list, "s": seed_list, "b": block_list}
        self.conf_rw.write_json(path, res)

    def extract_hashes(self, rpc_block_list):
        return list(map(lambda x: x["hash"], rpc_block_list))

    def extract_unique_seeds(self, rpc_block_list):
        unique_seeds = {
            x["account_data"]["source_seed"]
            for x in rpc_block_list
            if x["account_data"]["source_seed"] is not None
        }
        return sorted(unique_seeds)

    def extract_blocks(self, rpc_block_list):
        return list(map(lambda x: x["block"], rpc_block_list))

    def process_rpc_block_list(self, rpc_block_list):
        BlockAsserts.assert_blockgen_succeeded(rpc_block_list)
        return {
            "hash_list": self.extract_hashes(rpc_block_list),
            "seed_list": self.extract_unique_seeds(rpc_block_list),
            "block_list": self.extract_blocks(rpc_block_list)
        }

    def is_nested_list(self, rpc_block_list):
        return any(isinstance(i, list) for i in rpc_block_list[:2])



from .interface import DockerInterface
from .autoheal import DockerAutoHeal
import subprocess
import json

from typing import List, Optional
from nanomock.internal.utils import subprocess_run_capture_output


#Holds all the shared os independant implementations
class DockerMixin(DockerInterface):

    def __init__(self, compose_path: str, project_name: str):
        self.compose_path = compose_path
        self.project_name = project_name

    @staticmethod
    def _get_stop_flag(force_stop: bool) -> str:
        return "-t 0" if force_stop else ""

    def _get_docker_compose_command(self,
                                    command,
                                    nodes: Optional[List[str]] = None):
        base_command = [
            "docker-compose", "-f",
            str(self.compose_path), "-p", self.project_name
        ]
        base_command.extend(command)
        if nodes: base_command.extend(nodes)
        return " ".join(base_command)

    def _run_command(self, base_command, auto_heal=True):
        try:
            subprocess_run_capture_output(base_command, True, None)
        except subprocess.CalledProcessError as exc:
            if auto_heal:
                docker_healer = DockerAutoHeal()
                response = docker_healer.retry_heal(exc, True, None)
                return response
            raise exc

    def restart_container(self,
                          container_name: str,
                          command_interval: int = 0,
                          force_stop=True) -> bool:
        stop_flag = self._get_stop_flag(force_stop)
        subprocess_run_capture_output(
            f"docker stop {stop_flag} {container_name} && sleep {command_interval} && docker start {container_name}",
            shell=True)
        return True

    def stop_and_remove_container(self, container_name: str, force_stop=True):
        stop_flag = self._get_stop_flag(force_stop)
        subprocess_run_capture_output(
            f"docker stop {stop_flag} {container_name} && docker rm {container_name}",
            shell=True)

    def check_container_running(self, container_name: str):
        cmd = f"docker ps |grep {container_name}$ | wc -l"
        res = subprocess_run_capture_output(cmd)
        return res

    def compose_start(self, nodes: Optional[List[str]] = None):
        cmd = self._get_docker_compose_command(["up", "-d"], nodes)
        return self._run_command(cmd)

    def compose_restart(self, nodes: Optional[List[str]] = None):
        cmd = self._get_docker_compose_command(["restart"], nodes)
        return self._run_command(cmd)

    def compose_stop(self, nodes: Optional[List[str]] = None):
        cmd = self._get_docker_compose_command(["stop"], nodes)
        return self._run_command(cmd)

    def compose_down(self, nodes: Optional[List[str]] = None):
        cmd = self._get_docker_compose_command(["down"], nodes)
        return self._run_command(cmd)

    def compose_build(self):
        cmd = self._get_docker_compose_command(["build"])
        return self._run_command(cmd)

    def compose_pull(self):
        cmd = self._get_docker_compose_command(["pull"])
        return self._run_command(cmd)

    def container_count(self, container_names):
        result = subprocess.run(
            ["docker", "ps", "-a", "--format", "{{.Names}}"],
            stdout=subprocess.PIPE,
            text=True,
        )

        existing_container_names = result.stdout.strip().split("\n")

        count = 0
        for name in container_names:
            if name in existing_container_names:
                count += 1

        return count

    def create_network(self, network_name):
        try:
            subprocess.run(["docker", "network", "create", network_name],
                           check=True,
                           capture_output=True)
        except subprocess.CalledProcessError as exc:
            output = exc.stderr.decode()
            if exc.returncode == 1 and "already exists" in output:
                print(f"The network '{network_name}' already exists.")
            else:
                raise Exception(f"Error creating network: {exc}") from exc

    def get_network_gateway(self, network_name):
        try:
            completed_process = subprocess.run(
                ["docker", "network", "inspect", network_name],
                capture_output=True,
                check=True)
            output = completed_process.stdout.decode()
            data = json.loads(output)
            gateway = data[0]['IPAM']['Config'][0]['Gateway']
            return gateway
        except subprocess.CalledProcessError as e:
            raise Exception(f"Error inspecting network: {e}") from e
        except KeyError:
            raise Exception(
                "Unable to find the Gateway in the network's IPAM Config.")



import os
import subprocess
import tomli
import tomli_w
import oyaml as yaml
import secrets
import json
import copy

from datetime import datetime
from nanolib import Block
from extradict import NestedData
from pathlib import Path

from nanomock.modules.nl_nanolib import NanoLibTools, raw_high_precision_multiply
from nanomock.modules.nl_rpc import NanoRpc
from nanomock.internal.utils import read_from_package_if_needed, is_packaged_version, find_device_for_path, convert_to_bytes, get_mock_logger
from nanomock.internal.feature_toggle import toggle
from nanomock.docker import get_docker_interface_class


def str2bool(v):
    return str(v).lower() in ("yes", "true", "t", "1")


class ConfigReadWrite:

    @read_from_package_if_needed
    def read_json(self, path, is_packaged=False):
        with open(path, "r") as f:
            return json.load(f)

    @read_from_package_if_needed
    def read_file(self, path, is_packaged=False):
        with open(path, "r") as f:
            return f.readlines()

    @read_from_package_if_needed
    def read_toml(self, path, is_packaged=False):
        try:
            with open(path, "rb") as f:
                toml_dict = tomli.load(f)
                return toml_dict
        except tomli.TOMLDecodeError as e:
            raise FileExistsError("Invalid config file! \n {}".format(str(e)))

    @read_from_package_if_needed
    def read_yaml(self, path, is_packaged=False):
        with open(path, 'r') as f:
            return yaml.safe_load(f)

    def write_json(self, path, json_dict):
        with open(path, "w") as f:
            json.dump(json_dict, f, indent=2)

    def append_json(self, path, json_dict):
        with open(path, "a") as f:
            json.dump(json_dict, f)
            f.write('\n')

    def write_list(self, path, list):
        with open(path, "w") as f:
            print(*list, sep="\n", file=f)

    def append_line(self, path, line):
        with open(path, 'a') as file:
            file.write(line)

    def write_toml(self, path, content):
        with open(path, "wb") as f:
            tomli_w.dump(content, f)

    def write_yaml(self, path, content):
        with open(path, 'w') as f:
            yaml.dump(json.loads(str(content).replace("'", '"')),
                      f,
                      default_flow_style=False)


class ConfigParser:

    preconfigured_peers = []

    def __init__(self, app_dir, config_file, logger=None):
        self.logger = logger or get_mock_logger()

        self.enabled_services = []
        self._set_path_variables(app_dir, config_file)
        self.conf_rw = ConfigReadWrite()
        self.nano_lib = NanoLibTools()
        self.config_dict = self.conf_rw.read_toml(self.nl_config_path)
        self.compose_dict = self.conf_rw.read_yaml(self.default_compose_path,
                                                   is_packaged=True)
        self.__config_dict_add_genesis_to_nodes()
        self.__config_dict_set_node_variables()  #modifies config_dict
        self.__config_dict_set_default_values()  #modifies config_dict
        self.__set_preconfigured_peers()
        self.__set_node_accounts()
        self.__set_balance_from_vote_weight()
        self.__set_special_account_data()
        #self.__set_docker_compose()

    def _set_path_variables(self, app_dir, config_file):
        user_app_dir = Path(app_dir).resolve()

        if is_packaged_version():
            self.services_dir = "nanomock.internal.data.services"
            self.default_compose_path = f"{self.services_dir}.default_docker-compose.yml"
            self.default_nanomonitor_config = f"{self.services_dir}.nanomonitor.default_config.php"
            #self.default_nanocap_config = f"{self.services_dir}.nanocap.nanocap.config"
            #self.default_nanocap_compose = f"{self.services_dir}.nanocap.nanocap-compose.yml"
            self.default_nanomonitor_config = f"{self.services_dir}.nanovotevisu.default_docker-compose.yml"
        else:
            self.services_dir = Path().resolve(
            ) / "app" / "internal" / "data" / "services"

            self.default_compose_path = Path(
                self.services_dir) / "default_docker-compose.yml"
            self.default_nanomonitor_config = Path(
                self.services_dir) / "nanomonitor" / "default_config.php"
            self.default_nanovotevisu_config = Path(
                self.services_dir
            ) / "nanovotevisu" / "default_docker-compose.yml"

        self.nl_config_path = user_app_dir / config_file
        self.nano_nodes_path = user_app_dir / "nano_nodes"
        self.nodes_dir = self.nano_nodes_path / "{node_name}"
        self.compose_out_path = self.nano_nodes_path / "docker-compose.yml"
        self.tcp_analyzer_path = Path("tcp_analyzer")

    def __set_node_accounts(self):
        available_supply = 340282366920938463463374607431768211455 - int(
            self.config_dict.get("burn_amount", 0)) - 1
        for node in self.config_dict["representatives"]["nodes"]:

            if "key" in node:
                account_data = self.nano_lib.key_expand(node["key"])
            else:
                account_data = self.nano_lib.nanolib_account_data(
                    seed=node["seed"], index=0)

            node["account"] = account_data["account"]
            node["account_data"] = account_data
            if "vote_weight_percent" in node:
                node["balance"] = raw_high_precision_multiply(
                    available_supply, node["vote_weight_percent"])

    def __set_special_account_data(self):
        self.config_dict["burn_account_data"] = {
            "account":
            "nano_1111111111111111111111111111111111111111111111111111hifc8npp"
        }
        self.config_dict["genesis_account_data"] = self.nano_lib.key_expand(
            self.config_dict["genesis_key"])
        self.config_dict["canary_account_data"] = self.nano_lib.key_expand(
            self.config_dict["canary_key"])

    def __config_dict_set_node_variables(self):
        self.config_dict.setdefault("env", "local")
        modified_config = False

        if "remote_address" not in self.config_dict:
            self.config_dict["remote_address"] = '127.0.0.1'

        if "host_port_peer" not in self.config_dict["representatives"]:
            self.config_dict["representatives"]["host_port_peer"] = 44000
        if "host_port_peer" not in self.config_dict["representatives"]:
            self.config_dict["representatives"]["host_port_rpc"] = 45000
        if "host_port_peer" not in self.config_dict["representatives"]:
            self.config_dict["representatives"]["host_port_ws"] = 47000

        if "node_prefix" not in self.config_dict["representatives"]:
            self.config_dict["representatives"]["node_prefix"] = "ns"

        host_port_inc = 0  #set incremental ports for nodes starting with 0
        for node in self.config_dict["representatives"]["nodes"]:

            if "name" not in node:
                node["name"] = f"{secrets.token_hex(6)}".lower()
                self.logger.warning(
                    f'no name set for a node. New name : {node["name"]}')
                modified_config = True
            node["name"] = node["name"].lower()

            if "seed" not in node and not self.is_genesis(node):
                node["seed"] = secrets.token_hex(32)
                self.logger.warning(
                    f'no seed set for a node. New seed : {node["seed"]}')
                modified_config = True

            #Add ports for each node
            node["name"] = self.get_node_prefix() + node["name"]

            if self.config_dict["env"] == "gcloud": host_port_inc = 0

            node["host_port_peer"] = self.config_dict["representatives"][
                "host_port_peer"] + host_port_inc
            node["host_port_rpc"] = self.config_dict["representatives"][
                "host_port_rpc"] + host_port_inc
            node["host_port_ws"] = self.config_dict["representatives"][
                "host_port_ws"] + host_port_inc

            if "host_ip" not in node:
                node["host_ip"] = self.config_dict["remote_address"]

            node[
                "rpc_url"] = f'http://{node["host_ip"]}:{node["host_port_rpc"]}'
            node["ws_url"] = f'ws://{node["host_ip"]}:{node["host_port_ws"]}'
            host_port_inc = host_port_inc + 1

        if modified_config:
            user_input = "nl_config.toml was modified. Save current version? This will change the structure (y/n)"
            if user_input == 'y':
                self.conf_rw.write_toml(self.services_dir, self.config_dict)

    def __config_dict_set_default_values(self):
        #self.config_dict = conf_rw.read_toml(self.services_dir)
        self.config_dict["NANO_TEST_EPOCH_1"] = "0x000000000000000f"

        self.config_dict.setdefault(
            "genesis_key",
            "12C91837C846F875F56F67CD83040A832CFC0F131AF3DFF9E502C0D43F5D2D15")
        self.config_dict.setdefault(
            "canary_key",
            "FB4E458CB13508353C5B2574B82F1D1D61367F61E88707F773F068FF90050BEE")
        self.config_dict.setdefault("epoch_count", 2)
        self.config_dict.setdefault("NANO_TEST_EPOCH_2", "0xfff0000000000000")
        self.config_dict.setdefault("NANO_TEST_EPOCH_2_RECV",
                                    "0xfff0000000000000")
        self.config_dict.setdefault("NANO_TEST_MAGIC_NUMBER", "LC")
        self.config_dict.setdefault(
            "NANO_TEST_CANARY_PUB",
            "CCAB949948224D6B33ACE0E078F7B2D3F4D79DF945E46915C5300DAEF237934E")

        #nanolooker
        self.config_dict.setdefault(
            "nanolooker_enable",
            str2bool(self.config_dict.get("nanolooker_enable", False)))
        self.config_dict.setdefault("nanolooker_port", 42000)
        self.config_dict.setdefault("nanolooker_node_name", "genesis")
        self.config_dict.setdefault("nanolooker_mongo_port", 27017)

        #nanomonitor, nanoticker, nano-vote-visualizer
        self.config_dict.setdefault(
            "nanomonitor_enable",
            str2bool(self.config_dict.get("nanomonitor_enable", False)))
        self.config_dict.setdefault(
            "nanoticker_enable",
            str2bool(self.config_dict.get("nanoticker_enable", False)))
        self.config_dict.setdefault(
            "nanovotevisu_enable",
            str2bool(self.config_dict.get("nanovotevisu_enable", False)))

        #prom-exporter
        self.config_dict.setdefault(
            "promexporter_enable",
            str2bool(self.config_dict.get("promexporter_enable", False)))
        self.config_dict.setdefault(
            "prom_gateway",
            str2bool(
                self.config_dict.get("prom_gateway", "nl_pushgateway:9091")))
        self.config_dict.setdefault("prom_runid", "default")

        #traffic control
        self.config_dict.setdefault(
            "tc_enable", str2bool(self.config_dict.get("tc_enable", False)))

        #privileged
        self.config_dict.setdefault(
            "privileged", str2bool(self.config_dict.get("privileged", False)))

        #nanocap
        self.config_dict.setdefault(
            "nanocap_enable",
            str2bool(self.config_dict.get("nanocap_enable", False)))

        #tcpdump
        self.config_dict.setdefault(
            "tcpdump_enable",
            str2bool(self.config_dict.get("tcpdump_enable", False)))
        self.config_dict.setdefault(
            "tcpdump_filename",
            f"nl_tcpdump_{datetime.now().strftime('%Y%m%d_%H%M%S')}.pcap")

    def __config_dict_add_genesis_to_nodes(self):
        genesis_node_name = "genesis"
        genesis_node = next(
            (d for d in self.config_dict["representatives"]["nodes"]
             if d["name"] == genesis_node_name), None)

        if genesis_node:
            genesis_node.setdefault("key", self.config_dict["genesis_key"])
            genesis_node.setdefault("is_genesis", True)
            return

        self.config_dict["representatives"]["nodes"].insert(
            0, {
                "name": genesis_node_name,
                "key": self.config_dict["genesis_key"],
                "is_genesis": True
            })

    def __set_preconfigured_peers(self):
        for node in self.config_dict["representatives"]["nodes"]:
            if node["name"] not in self.preconfigured_peers:
                self.preconfigured_peers.append(node["name"])
        return self.preconfigured_peers

    def __is_principal_representative(self, available_supply, balance):
        response = False
        if int(balance) >= int(available_supply / 1000):
            response = True
        return response

    def __set_balance_from_vote_weight(self):
        available_supply = 340282366920938463463374607431768211455 - int(
            self.config_dict.get("burn_amount", 0)) - 1
        genesis_balance = available_supply
        for node_conf in self.get_nodes_config():
            if "vote_weight" in node_conf:
                node_conf["balance"] = raw_high_precision_multiply(
                    available_supply, node_conf["vote_weight"])
            #evaluate if a node is a PR
            #if a node has more than 0.1% of available supply it's considered a PR.
            #this is accurate in the case that all nodes are online. (which is reasonable in a private network)

            if "balance" in node_conf:
                genesis_balance = genesis_balance - int(node_conf["balance"])
                node_conf["is_pr"] = self.__is_principal_representative(
                    available_supply, node_conf["balance"])

        #add genesis_balance to config
        self.get_genesis_config(
        )["is_pr"] = self.__is_principal_representative(
            available_supply, genesis_balance)

    def value_in_dict(self, dict, value_l):
        for key, value in dict.items():
            if value == value_l:
                return {"found": True, "value": dict}
        return {"found": False, "value": None}

    def value_in_dict_array(self, dict_array, value_l):
        for dict in dict_array:
            dict_found = self.value_in_dict(dict, value_l)
            if dict_found["found"]:
                return dict_found
        return {"found": False, "value": None}

    def is_genesis(self, node_conf):
        if "is_genesis" in node_conf and node_conf["is_genesis"]:
            return True
        return False

    def is_voting_enabled(self, node_name):
        node_conf = self.get_node_config(node_name)
        if node_conf:
            return node_conf.get("enable_voting", True)
        raise ValueError(
            f"{node_name} undefined in {self.nl_config_path}\nValid names:{self.get_nodes_name()}"
        )

    def _remove_keys_with_value(self, d, k, val):
        #if dict (d), search for key (k) that has value (val) and remove key-value pairfrom dict
        if k in d and d[k] == val:
            d.pop(k)
        for v in d.values():
            if isinstance(v, list):
                for el in v:
                    self._remove_keys_with_value(el, k, val)
            elif isinstance(v, dict):
                return self._remove_keys_with_value(v, k, val)
        return None

    def modify_nanolocal_config(self,
                                nested_path: str,
                                nested_value: str,
                                save: bool = True):
        config_nested = NestedData(self.conf_rw.read_toml(self.nl_config_path))

        if nested_value is None:
            config_nested.merge("DELETE_ME", nested_path)
        else:
            config_nested.merge(nested_value, nested_path)

        self._remove_keys_with_value(config_nested.data,
                                     nested_path.split(".")[-1:][0],
                                     "DELETE_ME")

        if save:
            self.config_dict = config_nested.data
            self.conf_rw.write_toml(self.nl_config_path, config_nested.data)

        return config_nested

    def get_config_reader_writer(self):
        return self.conf_rw

    def set_prom_runid(self, runid):
        self.runid = runid

    def get_name_with_prefix(self, node_name):
        return self.get_node_prefix() + node_name

    def get_node_prefix(self):
        #set during initialisation in __config_dict_set_node_variables
        return self.config_dict["representatives"]["node_prefix"] + "_"

    def get_project_name(self):
        return self.get_node_prefix() + "nanomock"

    def get_xnolib_localctx(self):
        ctx = {
            'peers': {
                node_conf["name"]: {
                    "ip": f'::ffff:{node_conf["host_ip"]}',
                    "port": node_conf["host_port_peer"],
                    "score": 1000,
                    "is_voting": node_conf["is_pr"]
                }
                for node_conf in self.get_nodes_config()
            },
            'repservurl': '',
            'genesis_pub': self.get_genesis_account_data()["public"],
            'epoch_v2_signing_account':
            self.config_dict["NANO_TEST_CANARY_PUB"],
            'genesis_block': self.get_genesis_block(as_json=True),
            'peerserviceurl': ''
        }

        return ctx

    def get_canary_pub_key(self):
        env = self.config_dict["env"]
        canary_pub = ""
        if env in ["gcloud", "local"]:
            canary_pub = self.nano_lib.key_expand(
                self.config_dict["canary_key"])["public"]
        elif env == "beta":
            canary_pub = "868C6A9F79D4506E029B378262B91538C5CB26D7C346B63902FFEB365F1C1947"
        elif env == "live":
            canary_pub = "7CBAF192A3763DAEC9F9BAC1B2CDF665D8369F8400B4BC5AB4BA31C00BAA4404"
        else:
            raise ValueError(
                f'"{env}" is not in the list of accepted valued ["local", "beta", "live"] for variable "env" in nl_config.toml'
            )
        return canary_pub

    def get_network_name(self):
        return self.compose_dict["networks"]["nano-local"]["name"]

    def get_genesis_block(self, as_json=False):

        env = self.config_dict["env"]

        if env in ["gcloud", "local"]:
            genesis_account = self.get_genesis_account_data()
            block = Block(block_type="open",
                          account=genesis_account["account"],
                          representative=genesis_account["account"],
                          source=genesis_account["public"])

            block.solve_work(
                difficulty=self.config_dict["NANO_TEST_EPOCH_1"].replace(
                    "0x", ""))

            private_key = genesis_account["private"]
            block.sign(private_key)
            json_block = block.json()

        elif env == "beta":
            json_block = str({
                "type":
                "open",
                "source":
                "259A43ABDB779E97452E188BA3EB951B41C961D3318CA6B925380F4D99F0577A",
                "representative":
                "nano_1betagoxpxwykx4kw86dnhosc8t3s7ix8eeentwkcg1hbpez1outjrcyg4n1",
                "account":
                "nano_1betagoxpxwykx4kw86dnhosc8t3s7ix8eeentwkcg1hbpez1outjrcyg4n1",
                "work":
                "79d4e27dc873c6f2",
                "signature":
                "4BD7F96F9ED2721BCEE5EAED400EA50AD00524C629AE55E9AFF11220D2C1B00C3D4B3BB770BF67D4F8658023B677F91110193B6C101C2666931F57046A6DB806"
            }).replace("'", '"')

        elif env == "live":
            json_block = str({
                "type":
                "open",
                "source":
                "E89208DD038FBB269987689621D52292AE9C35941A7484756ECCED92A65093BA",
                "representative":
                "xrb_3t6k35gi95xu6tergt6p69ck76ogmitsa8mnijtpxm9fkcm736xtoncuohr3",
                "account":
                "xrb_3t6k35gi95xu6tergt6p69ck76ogmitsa8mnijtpxm9fkcm736xtoncuohr3",
                "work":
                "62f05417dd3fb691",
                "signature":
                "9F0C933C8ADE004D808EA1985FA746A7E95BA2A38F867640F53EC8F180BDFE9E2C1268DEAD7C2664F356E37ABA362BC58E46DBA03E523A7B5A19E4B6EB12BB02"
            }).replace("'", '"')
        else:
            raise ValueError(
                f'"{env}" is not in the list of accepted valued ["local", "beta", "live"] for variable "env" in nl_config.toml'
            )

        if as_json:
            return json.loads(json_block)

        return json_block

    def get_node_rpc(self, node_name):
        node_conf = self.get_node_config(node_name)
        if node_conf:
            return node_conf["rpc_url"]
        raise ValueError(
            f"{node_name} undefined in {self.nl_config_path}\nValid names:{self.get_nodes_name()}"
        )

    def get_nodes_rpc(self):
        api = []
        for node_name in self.get_nodes_name():
            node_conf = self.get_node_config(node_name)
            api.append(node_conf["rpc_url"])
        return api

    def get_nodes_rpc_port(self):
        api = {}
        for node_name in self.get_nodes_name():
            node_conf = self.get_node_config(node_name)
            api[node_name] = node_conf["rpc_url"].split(":")[2]
        return api

    def get_node_name_from_rpc_url(self, rpc_endpoint: NanoRpc):

        for node_name in self.get_nodes_name():
            node_conf = self.get_node_config(node_name)
            if rpc_endpoint.RPC_URL == node_conf["rpc_url"]: return node_name
        return None

    def get_remote_address(self):
        return self.config_dict["remote_address"]

    # def key_expand(self, private_key):

    #     signing_key = SigningKey(unhexlify(private_key))
    #     private_key = signing_key.to_bytes().hex()
    #     public_key = signing_key.get_verifying_key().to_bytes().hex()

    #     return {
    #         "private": private_key,
    #         "public": public_key,
    #         "account": self.h.public_key_to_nano_address(unhexlify(public_key))
    #     }

    def write_nanomonitor_config(self, node_name):
        nanomonitor_config = self.conf_rw.read_file(
            self.default_nanomonitor_config, is_packaged=True)
        destination_path = str(
            os.path.join(
                self.nodes_dir,
                "nanoNodeMonitor/config.php")).format(node_name=node_name)
        node_config = self.get_node_config(node_name)
        nanomonitor_config[4] = f"$nanoNodeName = '{node_name}';"
        nanomonitor_config[5] = f"$nanoNodeRPCIP   = '{node_name}';"
        nanomonitor_config[
            7] = f"$nanoNodeAccount = '{node_config['account']}';"
        self.conf_rw.write_list(destination_path, nanomonitor_config)

    def get_all(self):
        return self.config_dict

    def get_genesis_node_name(self):
        return self.config_dict["representatives"]["nodes"][0]["name"]

    def get_genesis_config(self):
        genesis_name = self.get_genesis_node_name()
        return self.get_node_config(genesis_name)

    def get_node_config(self, node_name):
        result = self.value_in_dict_array(
            self.config_dict["representatives"]["nodes"], node_name)
        if result["found"]:
            return result["value"]

    def get_genesis_account_data(self):
        return self.config_dict["genesis_account_data"]

    def get_burn_account_data(self):
        return self.config_dict["burn_account_data"]

    def get_canary_account_data(self):
        return self.config_dict["canary_account_data"]

    def get_max_balance_key(self):
        #returns the privatekey for the node with the highest defined balance.
        nodes_conf = self.get_nodes_config()
        max_balance = max(
            int(x["balance"]) if "balance" in x else 0
            for x in self.get_nodes_config())
        node_conf = list(
            filter(lambda x: int(x.get("balance", 0)) == max_balance,
                   self.get_nodes_config()))
        return node_conf[0]["account_data"]["private"]

    def get_nodes_name(self):
        response = []
        for node in self.config_dict["representatives"]["nodes"]:
            response.append(node["name"])
        return response

    def get_conatiners_name(self):
        if not os.path.exists(self.compose_out_path):
            return []

        compose_config = self.conf_rw.read_yaml(self.compose_out_path)
        return [service for service in compose_config["services"].keys()]

    def get_nodes_config(self):
        res = []
        for node_name in self.get_nodes_name():
            #res[node_name] = self.get_node_config(node_name)
            res.append(self.get_node_config(node_name))
        return res

    def set_node_balance(self, node_name, balance):
        self.get_node_config(node_name)["balance"] = balance

    def get_docker_compose_env_variables(self):
        conf_variables = self.config_dict
        env_variables = []
        genesis_block = json.loads(self.get_genesis_block())
        s_genesis_block = str(genesis_block).replace("'", '"')

        # Set genesis block
        env_variables.append(f"NANO_TEST_GENESIS_BLOCK={s_genesis_block}")

        env_variables.append(
            f'NANO_TEST_GENESIS_PUB="{genesis_block["source"]}"')
        env_variables.append(
            f'NANO_TEST_CANARY_PUB="{self.get_canary_pub_key()}"')

        for key, value in conf_variables.items():
            if key.startswith("NANO_TEST_"):
                env_variables.append(f'{key}="{value}"')

        return env_variables

    def set_network_name(self):
        self.compose_dict["networks"]["nano-local"][
            "name"] = self.get_node_prefix(
            ) + self.compose_dict["networks"]["nano-local"]["name"]

    def set_docker_compose(self):
        default_service_names = [
            service for service in self.compose_dict["services"]
        ]

        self.set_network_name()

        #Add nodes and ports
        for node in self.config_dict["representatives"]["nodes"]:
            self.compose_add_node(node["name"])
            self.compose_set_node_ports(node["name"])

        if self.get_config_value("nanolooker_enable"):
            self.set_nanolooker_compose()

        if self.get_config_value("nanomonitor_enable"):
            self.set_nanomonitor_compose()

        if self.get_config_value("nanoticker_enable"):
            self.set_nanoticker_compose()

        if self.get_config_value("nanovotevisu_enable"):
            self.set_nanovotevisu_compose()

        if bool(self.get_config_value("promexporter_enable")):
            self.set_promexporter_compose()

        if bool(self.get_config_value("tcpdump_enable")):
            self.set_tcpdump_compose()

        if bool(self.get_config_value("nanocap_enable")):
            self.set_nanocap_compose()

        #remove default container
        for service in default_service_names:
            self.compose_dict["services"].pop(service, None)

    def set_nanovotevisu_compose(self):
        nanoticker_compose = self.conf_rw.read_yaml(
            self.default_nanovotevisu_config)
        self.compose_dict["services"]["nl_nanovotevisu"] = nanoticker_compose[
            "services"]["nl_nanovotevisu"]
        self.compose_dict["services"]["nl_nanovotevisu"]["build"]["args"][
            0] = f'REMOTE_ADDRESS={self.get_config_value("remote_address")}'
        self.compose_dict["services"]["nl_nanovotevisu"]["build"]["args"][
            1] = f'HOST_ACCOUNT={self.get_node_config(self.get_nodes_name()[0])["account"]}'
        self.enabled_services.append(
            f'nano-vote-visualizer enabled at {self.get_config_value("remote_address")}:42001'
        )

    def set_nanoticker_compose(self):
        nanoticker_compose = self.conf_rw.read_yaml(
            f'{self.services_dir}/nanoticker/default_docker-compose.yml',
            is_packaged=True)
        self.compose_dict["services"]["nl_nanoticker"] = nanoticker_compose[
            "services"]["nl_nanoticker"]
        self.compose_dict["services"]["nl_nanoticker"]["build"]["args"][
            0] = f'REMOTE_ADDRESS={self.get_config_value("remote_address")}'
        self.enabled_services.append(
            f'nanoticker enabled at {self.get_config_value("remote_address")}:42002'
        )

    def set_nanolooker_compose(self):
        nanolooker_compose = self.conf_rw.read_yaml(
            f'{self.services_dir}/nanolooker/default_docker-compose.yml',
            is_packaged=True)

        for container in nanolooker_compose["services"]:
            container_name = self.get_node_prefix(
            ) + nanolooker_compose["services"][container]["container_name"]
            #Add all containers from docker-compose file to our compose_dict
            self.compose_dict["services"][container] = nanolooker_compose[
                "services"][container]
            #add prefix to container_name defined in docker-compose file

            self.compose_dict["services"][container][
                "container_name"] = container_name

        nanolooker_node_config = self.get_node_config(
            self.get_name_with_prefix(
                self.config_dict["nanolooker_node_name"]))

        #in webbrowser: access websocket of the remote machine instead of localhost
        self.compose_dict["services"]["nl_nanolooker"]["build"]["args"][
            0] = f'REMOTE_ADDRESS={self.get_config_value("remote_address")}'
        self.compose_dict["services"]["nl_nanolooker"]["build"]["args"][
            1] = f'MONGO_CONTAINER={self.compose_dict["services"]["nl_nanolooker_mongo"]["container_name"]}'
        self.compose_dict["services"]["nl_nanolooker"]["build"]["args"][
            2] = f'MONGO_PORT={self.config_dict["nanolooker_mongo_port"]}'
        self.compose_dict["services"]["nl_nanolooker"]["build"]["args"][
            3] = f'NODE_WEBSOCKET_PORT={nanolooker_node_config["host_port_ws"]}'
        #set node for RPC
        self.compose_dict["services"]["nl_nanolooker"]["environment"][
            2] = f'RPC_DOMAIN=http://{nanolooker_node_config["name"]}:17076'
        #set correct port
        self.compose_dict["services"]["nl_nanolooker"]["ports"][
            0] = f'{self.config_dict["nanolooker_port"]}:3010'
        self.enabled_services.append(
            f'nanolooker enabled at {self.get_config_value("remote_address")}:{self.config_dict["nanolooker_port"]}'
        )

    def set_nanomonitor_compose(self):
        host_port_inc = 0
        for node in self.config_dict["representatives"]["nodes"]:
            nanomonitor_compose = self.conf_rw.read_yaml(
                f'{self.services_dir}/nanomonitor/default_docker-compose.yml',
                is_packaged=True)
            container = nanomonitor_compose["services"]["default_monitor"]
            container_name = f'{node["name"]}_monitor'
            self.compose_dict["services"][container_name] = copy.deepcopy(
                container)
            self.compose_dict["services"][container_name][
                "container_name"] = container_name
            self.compose_dict["services"][container_name]["volumes"][
                0] = self.compose_dict["services"][container_name]["volumes"][
                    0].replace("default_monitor", node["name"])
            self.compose_set_nanomonitor_ports(container_name, host_port_inc)
            host_port_monitor = 46000 + host_port_inc
            self.enabled_services.append(
                f'nano-node-monitor enabled at {self.get_config_value("remote_address")}:{host_port_monitor}'
            )
            host_port_inc = host_port_inc + 1

    def set_promexporter_compose(self):
        DockerInterfaceClass = get_docker_interface_class()
        host_ip = DockerInterfaceClass.get_docker_gateway_ip()

        #Create prometheus, prom-gateway and grafana IF we use default prom-gateway
        if self.get_config_value("prom_gateway") == "nl_pushgateway:9091":
            promexporter_compose = self.conf_rw.read_yaml(
                f'{self.services_dir}/promexporter/default_docker-compose.yml',
                is_packaged=True)
            for container in promexporter_compose["services"]:
                self.compose_dict["services"][
                    container] = promexporter_compose["services"][container]
            for volume in promexporter_compose["volumes"]:
                self.compose_dict["volumes"][volume] = promexporter_compose[
                    "volumes"][volume]

            self.enabled_services.append(
                f'promgateway enabled at {self.get_config_value("remote_address")}:42005'
            )

        #Create 1 exporter per node
        for node in self.config_dict["representatives"]["nodes"]:
            node_config = self.get_node_config(node["name"])
            node_rpc_port = node_config["host_port_rpc"]

            prom_gateway = self.get_config_value("prom_gateway")
            prom_runid = self.get_config_value("prom_runid")

            nanomonitor_compose = self.conf_rw.read_yaml(
                f'{self.services_dir}/promexporter/default_exporter_docker-compose.yml',
                is_packaged=True)
            container = nanomonitor_compose["services"]["default_exporter"]
            container_name = f'{node["name"]}_exporter'
            self.compose_dict["services"][container_name] = copy.deepcopy(
                container)
            self.compose_dict["services"][container_name][
                "container_name"] = container_name
            self.compose_dict["services"][container_name]["command"]

            self.compose_dict["services"][container_name][
                "command"] = f'--rpchost {host_ip} --rpc_port {node_rpc_port} --push_gateway {prom_gateway} --hostname {node["name"]} --runid {prom_runid} --interval 2'

            self.compose_dict["services"][container_name][
                "pid"] = f'service:{node["name"]}'

            self.enabled_services.append(
                f'{container_name} added for node {node["name"]}')

    def set_nanocap_device_ip(self, device_ip):
        nanocap_config_path = self.nano_nodes_path / "services" / "nanocap" / "nanocap.config"
        nanocap_config = self.conf_rw.read_json(nanocap_config_path)
        nanocap_config["capture"]["device_ip"] = device_ip
        nanocap_config = self.conf_rw.write_json(nanocap_config_path,
                                                 nanocap_config)

    def set_nanocap_compose(self):

        #nanocap_config = self.conf_rw.read_json(self.default_nanocap_config,is_packaged=True)

        nanocap_compose = self.conf_rw.read_yaml(
            f'{self.services_dir}/nanocap/nanocap-compose.yml',
            is_packaged=True)
        container = nanocap_compose["services"]["nanocap"]
        container_name = f'{self.get_node_prefix()}nanocap'
        self.compose_dict["services"][container_name] = container
        self.compose_dict["services"][container_name][
            "container_name"] = container_name
        self.enabled_services.append(
            f'nanocap enabled ! This may lead to a decrease in performance!')

    def set_tcpdump_compose(self):

        tcp_analyzer_config_path = f'{self.tcp_analyzer_path}/config.json'
        if not os.path.exists(tcp_analyzer_config_path):
            conf_source_path = f'{self.services_dir}/tcpdump/tcp_analyzer_config.example.json'
            copy_conf = f'cp -p {conf_source_path} {tcp_analyzer_config_path}'
            os.system(copy_conf)

        tcp_analyzer_config = self.conf_rw.read_json(tcp_analyzer_config_path,
                                                     is_packaged=True)
        tcp_analyzer_config["files_name_in"] = []

        tcpdump_compose = self.conf_rw.read_yaml(
            f'{self.services_dir}/tcpdump/default_docker-compose.yml')
        container = tcpdump_compose["services"]["ns_tcpdump"]

        container_name = f'ns_tcpdump'
        pcap_file_name = f'{self.get_config_value("tcpdump_filename")}'

        #container_name = 'ns_tcpdump'

        self.compose_dict["services"][container_name] = container
        self.compose_dict["services"][container_name][
            "container_name"] = container_name

        #mount pcap file
        self.compose_dict["services"][container_name]["volumes"][
            0] = self.compose_dict["services"][container_name]["volumes"][
                0].replace("FILENAME", pcap_file_name)
        #network_mode
        self.compose_dict["services"][container_name]["network_mode"] = "host"

        #manually create the mounted file, otherwise docker-compose will create a directory
        pcap_file_path = f'{self.nano_nodes_path}/{pcap_file_name}'
        tcp_analyzer_config["files_name_in"].append(pcap_file_path)
        subprocess.call(f'touch {pcap_file_path}', shell=True)
        self.enabled_services.append(
            f'TCPDUMP enabled ! This may lead to a decrease in performance!')
        self.conf_rw.write_json(tcp_analyzer_config_path, tcp_analyzer_config)

    def get_enabled_services(self):
        return self.enabled_services

    def print_enabled_services(self):
        for service in self.enabled_services:
            self.logger.info(service)

    def get_config_value(self, key):
        if key not in self.config_dict:
            return None
        return self.config_dict[key]

    def write_docker_compose(self):
        self.conf_rw.write_yaml(self.compose_out_path, self.compose_dict)

    def get_config_tag(self, tag, node_name, default):
        #takes the first non empty tag.
        #First looks for the individual tag
        #then for the general tag
        #last uses the default
        individual_tag = self.get_representative_config(tag, node_name)
        general_tag = self.get_representative_config(tag, None)

        if individual_tag["found"]:
            return individual_tag["value"]
        elif general_tag["found"]:
            return general_tag["value"]
        else:
            return default

    def get_docker_tag(self, node_name):
        self.get_config_tag(self, "docker_tag", node_name,
                            "nanocurrency/nano-beta:latest")

    def get_disk_defaults(self, disk_type):
        disk_defaults = {
            "NVME": {
                "device_read_bps": "2000MB",
                "device_write_bps": "1000MB",
                "device_read_iops": "200000",
                "device_write_iops": "200000",
            },
            "SSD": {
                "device_read_bps": "400MB",
                "device_write_bps": "300MB",
                "device_read_iops": "50000",
                "device_write_iops": "40000",
            },
            "HDD": {
                "device_read_bps": "50MB",
                "device_write_bps": "50MB",
                "device_read_iops": "100",
                "device_write_iops": "100",
            },
        }
        return disk_defaults.get(disk_type.upper(), None)

    def add_container_blkio_config(self, container, node_name):
        if toggle.is_feature_disabled("config_blkio"): return
        blkio_config = {}
        config_tags = [
            "device_read_bps",
            "device_write_bps",
            "device_read_iops",
            "device_write_iops",
        ]

        disk_type = self.get_config_tag("disk", node_name, None)
        if disk_type:
            disk_defaults = self.get_disk_defaults(disk_type)
            if disk_defaults:
                for tag in config_tags:
                    rate = convert_to_bytes(disk_defaults[tag])
                    blkio_config[tag] = [{
                        "path":
                        find_device_for_path(self.nl_config_path),
                        "rate":
                        rate
                    }]
        else:
            for tag in config_tags:
                rate = self.get_config_tag(tag, node_name, None)
                if rate is not None:
                    blkio_config[tag] = [{
                        "path":
                        find_device_for_path(self.nl_config_path),
                        "rate":
                        convert_to_bytes(rate)
                    }]

        if blkio_config:
            container["blkio_config"] = blkio_config

    def add_container_cpu_memory_config(self, container, node_name):
        cpu = self.get_config_tag("cpu", node_name, None)
        if cpu is not None:
            container["cpus"] = float(cpu)

        memory = self.get_config_tag("memory", node_name, None)
        if memory is not None:
            container["mem_limit"] = convert_to_bytes(memory)

    def get_container_type(self, user_id):
        if user_id == "0":
            return "default_docker_root"
        elif user_id == "1000":
            return "default_docker"
        else:
            return "default_docker_custom"

    def set_container_image_or_build_args(self, container, user_id,
                                          docker_tag):
        if user_id in ["0", "1000"]:
            container["image"] = f"{docker_tag}"
        else:
            container["build"]["args"] = [
                f'NANO_IMAGE={docker_tag}',
                f'UID={user_id}',
            ]

    def compose_add_node(self, node_name):
        user_id = str(os.getuid())
        docker_tag = self.get_config_tag("docker_tag", node_name,
                                         "nanocurrency/nano-beta:latest")

        container_type = self.get_container_type(user_id)
        container = self.compose_add_container(node_name, container_type)
        if self.config_dict["privileged"]:
            container["privileged"] = 'true'

        if user_id != "1000" or self.config_dict["tc_enable"]:
            container["user"] = user_id

        if self.config_dict["tc_enable"]:
            container["build"]["args"].append(
                f'TC_ENABLE={str(self.config_dict["tc_enable"]).upper()}')

        self.set_container_image_or_build_args(container, user_id, docker_tag)
        if container:
            self.add_container_blkio_config(container, node_name)
            self.add_container_cpu_memory_config(container, node_name)

    def compose_set_node_ports(self, node_name):
        node_config = self.get_node_config(node_name)
        self.compose_dict["services"][node_name]["ports"] = [
            f'{node_config["host_port_peer"]}:17075',
            f'{node_config["host_port_rpc"]}:17076',
            f'{node_config["host_port_ws"]}:17078'
        ]

    def compose_set_nanomonitor_ports(self, container_name, port_i):
        host_port_monitor = 46000 + port_i
        self.compose_dict["services"][container_name]["ports"] = [
            f'{host_port_monitor}:80'
        ]

    def cp_dockerfile_and_nano_node(self, exec_path, node_name):
        #copy nano_node into working directory for Dockerfile
        dockerfile_path = self.nodes_dir.format(node_name=node_name)
        if exec_path.split(".")[-1] == "deb":
            copy_node = f"cp -p {exec_path} {dockerfile_path}/package.deb"
            copy_dockerfile = f"cp -p {self.services_dir}/default_deb_Dockerfile {dockerfile_path}/Dockerfile"
        else:
            copy_node = f"cp -p {exec_path} {dockerfile_path}/nano_node"
            copy_dockerfile = f"cp -p {self.services_dir}/default_Dockerfile {dockerfile_path}/Dockerfile"
        if os.path.exists(exec_path):
            os.system(copy_node)
            os.system(copy_dockerfile)
        else:
            self.logger.error(
                f'No nano_node could be found at [{exec_path}]. This container will fail on start'
            )

        return dockerfile_path

    def compose_add_container(self, node_name, default_container):
        #copies a default container and adds it as a new container
        self.compose_dict["services"][node_name] = copy.deepcopy(
            self.compose_dict["services"][default_container])
        self.compose_dict["services"][node_name]["container_name"] = node_name
        self.compose_dict["services"][node_name]["volumes"][
            0] = self.compose_dict["services"][node_name]["volumes"][
                0].replace("${default_docker}", node_name)
        return self.compose_dict["services"][node_name]

    def get_config_from_path(self, node_name, config_path_key):
        #returns None if no path is found
        config_dict_l = None
        if self.get_representative_config(
                config_path_key,
                node_name)["found"]:  #search by individual path
            config_dict_l = self.conf_rw.read_toml(
                self.get_representative_config(config_path_key,
                                               node_name)["value"])
        elif self.get_representative_config(
                config_path_key, None)["found"]:  #search by shared path
            config_dict_l = self.conf_rw.read_toml(
                self.get_representative_config(config_path_key, None)["value"])
        else:
            pass  #return None
        return config_dict_l

    def get_representative_config(self, node_key, node_name):
        #scan node config and match by name. Return the value of the key found in the config
        #response : {"found" : Bool, "value" = ...}
        if node_name is None and node_key is None:
            return {"found": False}

        if node_name is None:
            #shared config
            if node_key in self.config_dict["representatives"]:
                return {
                    "found": True,
                    "value": self.config_dict["representatives"][node_key]
                }
        else:
            #individual config
            representatives_config = self.value_in_dict_array(
                self.config_dict["representatives"]["nodes"], node_name)
            if representatives_config["found"] == True:
                if node_key in representatives_config["value"]:
                    return {
                        "found": True,
                        "value": representatives_config["value"][node_key]
                    }
        return {"found": False}


import os
import json
import subprocess
from typing import List


class DependencyChecker:

    def __init__(self):
        self.dependencies = self._load_dependencies_config()

    @staticmethod
    def _load_dependencies_config():
        return {
            "docker": ["docker", "--version"],
            "docker-compose": ["docker-compose", "--version"]
        }

        # config_file = "dependencies_config.json"
        # with open(config_file, "r") as f:
        #     return json.load(f)

    def check_dependencies(self):
        missing_dependencies = []
        for dep, command in self.dependencies.items():
            if not self._is_dependency_installed(command):
                missing_dependencies.append(dep)

        if missing_dependencies:
            self._display_missing_dependencies_instructions(
                missing_dependencies)
            return False
        return True

    @staticmethod
    def _is_dependency_installed(command: str) -> bool:
        try:
            subprocess.run(command,
                           check=True,
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL)
            return True
        except subprocess.CalledProcessError:
            return False

    @staticmethod
    def _display_missing_dependencies_instructions(
            missing_dependencies: List[str]):
        print("The following dependencies are missing:")
        for dep in missing_dependencies:
            print(f"- {dep}")

        print("\nPlease install the missing dependencies and try again.")


if __name__ == "__main__":
    checker = DependencyChecker()
    checker.check_dependencies()



from nanolib import Block, get_account_id, get_account_key_pair, AccountIDPrefix, generate_account_private_key, get_account_public_key
from decimal import Decimal, getcontext


def raw_high_precision_multiply(raw, multiplier) -> int:
    getcontext().prec = 1000  # Adjust precision as needed
    raw_amount = (Decimal(str(raw)) * Decimal(str(multiplier)))
    return int(raw_amount)


def raw_high_precision_percent(raw, percent) -> int:
    getcontext().prec = 1000  # Adjust precision as needed
    raw_amount = (Decimal(str(raw)) * Decimal(str(percent)) / Decimal('100'))
    return int(raw_amount)


class NanoLibTools():

    def get_account_from_public(self, public_key):
        return get_account_id(public_key=public_key,
                              prefix=AccountIDPrefix.NANO)

    def key_expand(self, private_key):
        account_key_pair = get_account_key_pair(private_key)
        account = get_account_id(public_key=account_key_pair.public,
                                 prefix=AccountIDPrefix.NANO)
        response = {
            "private": account_key_pair.private,
            "public": account_key_pair.public,
            "account": account
        }
        return response

    def nanolib_account_data(self, private_key=None, seed=None, index=0):
        if seed is not None:
            private_key = generate_account_private_key(seed, index)
        response = self.key_expand(private_key)

        if seed is not None:
            response["seed"] = seed
            response["index"] = index
        return response

    def get_state_block(self, account, representative, previous, balance,
                        link):

        return Block(block_type="state",
                     account=account,
                     representative=representative,
                     previous=previous,
                     balance=balance,
                     link=link)

    def create_state_block(self,
                           account,
                           representative,
                           previous,
                           balance,
                           link,
                           key,
                           difficulty=None):

        block = self.get_state_block(account, representative, previous,
                                     balance, link)
        block.solve_work(difficulty=difficulty)
        block.sign(key)
        return block



from abc import ABC, abstractmethod
from typing import List, Optional

class OSOperationsInterface(ABC):

    @abstractmethod
    def reset_nodes_data(self, nodes: Optional[List[str]] = None):
        pass

    # Add more method signatures as needed



from abc import ABC, abstractmethod
from typing import List, Optional


class DockerInterface(ABC):

    def __init__(self, compose_path: str, project_name: str):
        self.compose_path = compose_path
        self.project_name = project_name

    @classmethod
    @abstractmethod
    def get_docker_gateway_ip(cls) -> str:
        pass

    @abstractmethod
    def stop_and_remove_container(self, container_name: str, force_stop=True):
        pass

    @abstractmethod
    def check_container_running(self, container_name: str):
        pass

    @abstractmethod
    def compose_start(self, nodes: Optional[List[str]] = None):
        pass

    @abstractmethod
    def compose_restart(self, nodes: Optional[List[str]] = None):
        pass

    @abstractmethod
    def compose_stop(self, nodes: Optional[List[str]] = None):
        pass

    @abstractmethod
    def compose_down(self, nodes: Optional[List[str]] = None):
        pass

    @abstractmethod
    def compose_build(self):
        pass

    @abstractmethod
    def compose_pull(self):
        pass

    @abstractmethod
    def container_count(self, container_names):
        pass

    @abstractmethod
    def create_network(self, network_name):
        pass

    @abstractmethod
    def get_network_gateway(self, network_name):
        pass



import re
from nanomock.internal.utils import subprocess_run_capture_output, get_mock_logger
from subprocess import CalledProcessError


class DockerAutoHeal():

    logger = get_mock_logger()

    def __init__(self, max_heal_attempts=10):
        self.max_heal_attempts = max_heal_attempts

    def _get_error_mapping(self):
        error_mapping = {
            "address_in_use": {
                "error_msg": "programming external connectivity on endpoint",
                "heal_method": self._heal_address_in_use
            },
            "docker_in_use": {
                "error_msg":
                "Error response from daemon: Conflict. The container name",
                "heal_method": self._heal_docker_in_use
            },
            "file_not_found": {
                "error_msg": "No such file or directory",
                "heal_method": self._heal_file_not_found
            }
        }
        return error_mapping

    def heal_if_possible(self, error_msg, stderr):
        error_mapping = self._get_error_mapping()

        matching_errors = [
            error_data for error_data in error_mapping.values()
            if error_data["error_msg"] in stderr
        ]

        if matching_errors:
            healing_method = matching_errors[0]["heal_method"]
            return healing_method(error_msg, stderr)

        return False

    def _heal_file_not_found(self, error_msg, stderr):
        # Here, we are simply logging the error and returning True to indicate the error is healed.
        self.logger.warning(
            "Compose file not available. Nothing to take down.")
        return True

    def _heal_address_in_use(self, error_msg, stderr):
        error_msg = "programming external connectivity on endpoint"

        container_name = re.search(r"{} (\w+)".format(error_msg),
                                   stderr).group(1)
        subprocess_run_capture_output(
            f"docker stop -t 0 {container_name} && sleep 5 && docker start {container_name}",
            shell=True)
        return True

    def _heal_docker_in_use(self, error_msg, stderr):
        pattern = r'{} "/([^"]+)"'.format(error_msg)
        match = re.search(pattern, stderr)
        if match:
            container_name = match.group(1)
            subprocess_run_capture_output(
                f"docker stop -t 0 {container_name} && docker rm {container_name} && sleep 5",
                shell=True)
            return True
        return False

    def retry_heal(self, error: CalledProcessError, cmd_shell, cmd_cwd):
        stderr = error.stderr
        attempt = 1
        while attempt <= self.max_heal_attempts:
            if self.heal_if_possible(error.cmd, stderr):
                return True  # healing was successful
            self.logger.warning("Retry attempt %s\n%s", attempt, stderr)
            try:
                subprocess_run_capture_output(error.cmd, cmd_shell, cmd_cwd)
            except CalledProcessError as heal_error:
                self.logger.error("Attempt %s failed\n%s", attempt,
                                  heal_error.stderr)
                attempt += 1
        # healing was unsuccessful
        return error



import traceback
import requests
import json
import secrets
import time
import logging
from nanomock.modules.nl_nanolib import NanoLibTools, get_account_public_key

# for block_creation, we store local frontier info, so that subsequant calls know about the most recent frontier without needing to publish the block to the ledger.
_FRONTIER_INFO = {}
#introduced to reduce number of RPC calls for static values (network difficulty)
_GLOBAL_CACHE = {}


class NanoRpc:

    # api_config = None
    # debug = False

    def __init__(self, url, username=None, password=None):
        self.debug = True
        self.RPC_URL = url
        self.RPC_USER = username
        self.RPC_PASSWORD = password
        self.aio_conn = None  #aiohttp.TCPConnector(limit_per_host=100, limit=0, ttl_dns_cache=300)
        self.aio_results = []
        self.nano_lib = NanoLibTools()

    def clear_in_mem_account_info(self):
        _FRONTIER_INFO.clear()
        return _FRONTIER_INFO

    def request_get(self, url):
        try:
            headers = {
                "Content-type": "application/json",
                "Accept": "text/plain"
            }
            r = requests.get(url, headers=headers, timeout=1)
            response = {
                "status_code": r.status_code,
                "message": json.loads(r.text),
                "success": True
            }
        except Exception as e:
            response = {"success": False, "error_message": str(e)}
        return response

    def post_with_auth(self, content, max_retry=2, timeout=3, silent=True):
        try:
            url = self.RPC_URL
            headers = {
                "Content-type": "application/json",
                "Accept": "text/plain"
            }
            r = requests.post(url,
                              json=content,
                              headers=headers,
                              timeout=timeout)
            r_json = json.loads(r.text)

            # print("request: {} \rrepsonse: {}".format(content["action"], r.text ))ƒ
            if "error" in r_json:
                msg = "error in post_with_auth |\n request: \n{}\nresponse:\n{}".format(
                    content, r.text)
                if r_json["error"] == "Account not found":
                    logging.debug(msg)
                else:
                    if silent: logging.debug(msg)
                    else: logging.warn(msg)
            return r_json
        except Exception as e:
            if self.debug:
                logging.debug(
                    f'Error str{e} ... {max_retry} Retrys left for post_with_auth : {content["action"]}'
                )
            max_retry = max_retry - 1
            if max_retry >= 0:
                time.sleep(0.1)  #100ms
                self.post_with_auth(content, max_retry, timeout=timeout)

    def is_online(self, timeout=1):
        while timeout > 0:
            try:
                logging.debug("block_count: " +
                              self.block_count(max_retry=0)["count"])
                return True
            except:
                timeout = timeout - 1
                time.sleep(1)
        return False

    def generate_seed(self):
        return secrets.token_hex(32)

    def generate_new_seed(self):
        return {
            'success': True,
            'seed': self.generate_seed(),
            'error_message': ''
        }

    def validate_seed(self, seed):
        result = {'seed': seed, 'success': False, 'error_message': ''}
        if len(seed) == 64:
            try:
                int(seed, 16)
                result['success'] = True

            except Exception:
                result['error_message'] = 'Wrong seed format'
        else:
            result['error_message'] = 'Wrong seed length'

        return result

    def get_active_difficulty(self, request_only=False, use_cache=True):
        global _GLOBAL_CACHE
        req = {"action": "active_difficulty"}
        if request_only: return req
        if str(req) in _GLOBAL_CACHE and use_cache:
            result = _GLOBAL_CACHE[str(req)]
        else:
            result = self.post_with_auth(req)
            _GLOBAL_CACHE[str(req)] = result
        return result

    # def get_block_count(self ,request_only =False):
    #     req_active_difficulty = {"action" : "block_count"}
    #     return self.post_with_auth(req_active_difficulty)

    def get_account_data(self, seed, index=0):
        payload = self.generate_account(seed, index)
        payload["success"] = True
        payload["error_message"] = ''

        return payload

    def publish_blocks(self, blocks, json_data=True, sync=True):
        publish_commands = [{
            "action":
            "process",
            "json_block":
            "true",
            "subtype":
            block["subtype"],
            "block":
            (block if json_data else json.loads(block.replace("'", '"')))
        } for block in blocks]
        return self.__publish(payload_array=publish_commands,
                              json_data=json_data,
                              sync=sync)

    def publish_block(self, json_block, subtype=None):
        return self.__publish(
            json_block=json_block,
            subtype=subtype if subtype is not None else
            json_block["subtype"] if "subtype" in json_block else None)

    def send_publish_commands(self, publish_commands, sync=True):
        return self.__publish(payload_array=publish_commands,
                              json_data=True,
                              sync=sync)

    def __publish(self,
                  payload=None,
                  payload_array=None,
                  json_block=None,
                  subtype=None,
                  timeout=3,
                  sync=True,
                  json_data=False):
        if json_block is not None:
            if subtype is None:
                logging.warning(
                    "It's dangerous to publish blocks without subtype!")
                payload = {
                    "action": "process",
                    "json_block": "true",
                    "block": json_block,
                }
            else:
                payload = {
                    "action": "process",
                    "json_block": "true",
                    "subtype": subtype,
                    "block": json_block,
                }
        if payload is not None:
            return self.post_with_auth(json.loads(
                str(payload).replace("'", '"')),
                                       timeout=timeout)

    def block_hash(self, json_block, request_only=False):
        req = {
            "action": "block_hash",
            "json_block": "true",
            "block": json_block
        }
        if request_only: return req
        return self.post_with_auth(req)

    def block_info(self, block_hash, request_only=False):
        req = {
            "action": "block_info",
            "json_block": "true",
            "hash": block_hash
        }
        if request_only: return req
        return self.post_with_auth(req)

    def confirmation_active(self, request_only=False):
        req = {"action": "confirmation_active"}
        if request_only: return req
        resp = self.post_with_auth(req)
        if resp["confirmations"] == "": resp["confirmations"] = []
        return resp

    def block_confirmed(self, json_block=None, block_hash=None):
        if json_block is not None:
            block_hash = self.block_hash(json_block)["hash"]
        if block_hash is None:
            return False
        response = self.block_info(block_hash)
        if response is None:
            return False
        if "error" in response:
            return False
        return True if response["confirmed"] == "true" else False

    def block_count(self, max_retry=2, request_only=False):
        req = {"action": "block_count"}
        if request_only: return req
        resp = self.post_with_auth(req, max_retry=max_retry)
        return resp

    def get_stats(self, type="counters", request_only=False):
        req = {"action": "stats", "type": str(type)}
        if request_only: return req
        resp = self.post_with_auth(req)
        return resp

    def version(self, request_only=False):
        req = {"action": "version"}
        if request_only: return req
        resp = self.post_with_auth(req)
        return resp

    def generate_account(self, seed, index, request_only=False):
        req = {
            "action": "deterministic_key",
            "seed": seed,
            "index": index,
        }
        if request_only: return req
        account_data = self.post_with_auth(req)
        print("")

        account_data = {
            "seed":
            seed,
            "index":
            index,
            "private":
            account_data["private"],
            "public":
            account_data["public"],
            "account":
            account_data["account"],
            "nano_prefix":
            account_data["account"][0:11],
            "nano_center":
            account_data["account"][11:59],
            "nano_suffix":
            account_data["account"][len(account_data["account"]) - 6:]
        }
        return account_data

    def validate_account_number(self, account, request_only=False):
        response = {"success": False}
        req = {
            "action": "validate_account_number",
            "account": account,
        }
        if request_only: return req
        data = self.post_with_auth(req)
        if data["valid"] == "1":
            response["success"] = True
        return response

    def unlock_wallet(self, wallet, password, request_only=False):
        response = {"success": False}
        req = {
            "action": "password_enter",
            "wallet": wallet,
            "password": password
        }
        if request_only: return req
        data = self.post_with_auth(req)
        if data["valid"] == "1":
            response["success"] = True
        return response

    def wallet_create(self, seed, request_only=False):
        # response = {"success" : False}
        if seed == None:
            req = {"action": "wallet_create"}
        else:
            req = {
                "action": "wallet_create",
                "seed": seed,
            }
        if request_only: return req
        data = self.post_with_auth(req)
        # {
        #     "wallet": "646FD8B5940AB5B1AD2C0B079576A4CF5A25E8ADB10C91D514547EF5C10C05B7",
        #     "last_restored_account": "nano_3mcsrncubmquwcwgiouih17fjo8183t497c3q9w6qtnwz8bp3fig5x8m4rkw",
        #     "restored_count": "1"
        # }
        return data

    def wallet_add(self, wallet, private_key, request_only=False):
        # response = {"success" : False}
        req = {"action": "wallet_add", "wallet": wallet, "key": private_key}
        if request_only: return req
        data = self.post_with_auth(req)
        # {
        #   "account": "nano_3e3j5tkog48pnny9dmfzj1r16pg8t1e76dz5tmac6iq689wyjfpi00000000"
        # }
        return data

    def key_expand(self, private_key, request_only=False):
        req = {"action": "key_expand", "key": private_key}
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def peers(self, request_only=False):
        req = {"action": "peers"}
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def confirmation_quorum(self, request_only=False):
        req = {"action": "confirmation_quorum"}
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def account_info(self, account, request_only=False):
        req = {
            "action": "account_info",
            "account": account,
            "representative": "true",
            "pending": "true",
            "include_confirmed": "true"
        }
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def block_create_rpc(self,
                         balance,
                         account,
                         key,
                         representative,
                         link,
                         previous,
                         request_only=False):
        if previous is None: previous = "0" * 64
        req = {
            "action": "block_create",
            "json_block": "true",
            "type": "state",
            "balance": str(balance),
            "account": account,
            "key": key,
            "representative": representative,
            "link": link,
            "previous": previous,
            "difficulty": self.get_active_difficulty()["network_minimum"]
        }
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def block_create(self,
                     balance,
                     account,
                     key,
                     representative,
                     link,
                     previous,
                     request_only=False):

        block = self.nano_lib.create_state_block(
            account,
            representative,
            previous,
            balance,
            link,
            key,
            difficulty=self.get_active_difficulty()["network_minimum"])

        return {
            "hash": block.block_hash,
            "difficulty": block.difficulty,
            "block": json.loads(block.json())
        }

    def representatives_online(self, weight="false", request_only=False):
        req = {
            "action": "representatives_online",
            "weight": str(weight).lower()
        }
        if request_only: return req
        data = self.post_with_auth(req)
        return data

    def check_balance(self,
                      account,
                      include_only_confirmed=True,
                      request_only=False):

        multiplier = 10**30
        req = {
            "action": "account_balance",
            "account": account,
            "include_only_confirmed": include_only_confirmed
        }
        if request_only: return req
        data = self.post_with_auth(req)

        return {
            "account":
            account,
            "balance_raw":
            int(data["balance"]),
            "balance":
            self.truncate(int(data["balance"]) / multiplier),
            "pending":
            self.truncate(int(data["pending"]) / multiplier),
            "total":
            self.truncate(
                (int(data["balance"]) + int(data["pending"])) / multiplier)
        }

    def check_balances(self, seed, start_index=0, end_index=50):
        # check if there is any balance for account 0 to 50 accounts
        # {'index' : '' , 'account': '', 'balance': '', 'pending': ''}  ; spendable, total  Balance : 100 Nano . ! 95 Nano are currently not spendable. Your action is required.
        result = []
        for index in range(start_index, end_index + 1):
            nano_account = self.generate_account(seed, index)
            result.append(self.check_balance(nano_account["account"]))
        return result

    def truncate(self, number):
        if number > 0:
            return str('{:8f}'.format(number))
        else:
            return "0.00"

    def account_key(self, account, request_only=False):
        req = {"action": "account_key", "account": account}
        if request_only: return req
        data = self.post_with_auth(req)  #["key"]
        return data

    def get_pending_blocks(self, nano_account, threshold, number_of_blocks):

        response = {
            "account": nano_account,
            "blocks": None,
            "success": True,
            "error_message": ""
        }

        req_accounts_pending = {
            "action": "accounts_pending",
            "accounts": [nano_account],
            "threshold": str(threshold),
            "sorting": "true",
            "count": str(number_of_blocks)
        }
        accounts_pending = self.post_with_auth(req_accounts_pending)

        if "error" in accounts_pending:
            response["success"] = False
            response["error_message"] = accounts_pending["error"]
        elif accounts_pending["blocks"][nano_account] == "":
            response["success"] = False
            response["error_message"] = "no pending blocks"
        else:
            response["blocks"] = accounts_pending["blocks"][nano_account]

        return response

    def create_block(self,
                     sub_type,
                     link=None,
                     destination_account=None,
                     representative=None,
                     source_seed=None,
                     source_index=None,
                     source_private_key=None,
                     amount_raw=None,
                     in_memory=False,
                     add_in_memory=False,
                     read_in_memory=False,
                     use_rpc=True):
        try:
            if in_memory:
                add_in_memory = True
                read_in_memory = True
            if source_private_key is not None:
                source_account_data = self.nano_lib.nanolib_account_data(
                    private_key=source_private_key)
            elif source_seed is not None and source_index is not None:
                source_account_data = self.nano_lib.nanolib_account_data(
                    seed=source_seed, index=source_index)

            if read_in_memory:
                if source_account_data["account"] in _FRONTIER_INFO:
                    source_account_info = _FRONTIER_INFO[
                        source_account_data["account"]]
                else:
                    source_account_info = self.account_info(
                        source_account_data["account"])
            else:
                source_account_info = self.account_info(
                    source_account_data["account"])

            if representative is None:
                representative = source_account_info["representative"]
            if "balance" in source_account_info:
                balance = source_account_info["balance"]
            if "frontier" in source_account_info:
                previous = source_account_info["frontier"]

            if sub_type == "open" or sub_type == "receive":
                #destination_account = source_account_data["account"]
                if "error" in source_account_info:
                    sub_type = "open"
                    previous = None
                    balance = amount_raw
                    link = link
                else:
                    sub_type = "receive"
                    previous = source_account_info["frontier"]
                    balance = int(
                        source_account_info["balance"]) + int(amount_raw)
                    link = link

            elif sub_type == "send":
                link = get_account_public_key(account_id=destination_account)
                balance = int(source_account_info["balance"]) - int(amount_raw)
                previous = source_account_info["frontier"]

            elif sub_type == "change":
                amount_raw = "0"
                destination_account = source_account_data["account"]
                link = link

            elif sub_type == "epoch":
                if use_rpc: pass
                else: balance = int(source_account_info["balance"])

            if use_rpc:
                block = self.block_create_rpc(balance,
                                              source_account_data["account"],
                                              source_account_data["private"],
                                              representative, link, previous)
            else:
                block = self.block_create(balance,
                                          source_account_data["account"],
                                          source_account_data["private"],
                                          representative, link, previous)

            block["private"] = source_account_data["private"]
            block["subtype"] = sub_type
            block["amount_raw"] = amount_raw

            if "error" in block:
                block["success"] = False
                block["block"] = {}
                block["hash"] = None
            else:
                block["success"] = True
                block["error"] = None
                if add_in_memory:
                    _FRONTIER_INFO[source_account_data["account"]] = {
                        "frontier": block["hash"],
                        "balance": balance,
                        "representative": representative
                    }
            block["block"]["subtype"] = sub_type

        except Exception as e:
            traceback.print_exc()
            block = {
                "success": False,
                "block": {},
                "hash": None,
                "subtype": sub_type,
                "error": str(e)
            }
        return block

    def get_published_state(self, block, broadcast):
        published = False
        if broadcast:
            publish = self.publish_block(block["block"],
                                         subtype=block["subtype"])
            if "hash" in publish: published = True

    def get_block_result(self,
                         block,
                         broadcast,
                         source_seed=None,
                         source_index=None,
                         exit_after_s=2):
        start_time = time.time()
        if not block["success"]:
            logging.warning(block["error"])
        if broadcast:
            #with timeout(exit_after_s, exception=RuntimeError) :

            publish = None
            while publish is None:
                if time.time() - start_time > exit_after_s: break
                publish = self.publish_block(block["block"],
                                             subtype=block["subtype"])
                if publish is None:
                    time.sleep(0.5)
                    broadcast = False
                    logging.error(f'block not published : {block["hash"]}')
                    continue
                broadcast = True if "hash" in publish else False

        result = {
            "success":
            block["success"],
            "published":
            broadcast,
            "balance_raw":
            block["block"]["balance"] if "balance" in block["block"] else "",
            "amount_raw":
            block["amount_raw"] if "amount_raw" in block else "0",
            "hash":
            block["hash"],
            "block":
            block["block"],
            "subtype":
            block["subtype"],
            "account_data": {
                "account":
                block["block"]["account"]
                if "account" in block["block"] else "",
                "private":
                block["private"] if "private" in block else "",
                "source_seed":
                source_seed,
                "source_index":
                source_index
            },
            "error":
            block["error"]
        }

        if not result["success"]: print(result)

        return result

    def create_open_block(self,
                          destination_account,
                          open_private_key,
                          amount_per_chunk_raw,
                          rep_account,
                          send_block_hash,
                          broadcast=True):
        block = self.create_block("receive",
                                  source_private_key=open_private_key,
                                  destination_account=destination_account,
                                  representative=rep_account,
                                  amount_raw=amount_per_chunk_raw,
                                  link=send_block_hash,
                                  in_memory=not broadcast)

        return self.get_block_result(block, broadcast)

    def create_send_block(self,
                          source_seed,
                          source_index,
                          destination_account,
                          amount_per_chunk_raw,
                          broadcast=True):
        block = self.create_block("send",
                                  source_seed=source_seed,
                                  source_index=source_index,
                                  destination_account=destination_account,
                                  amount_raw=amount_per_chunk_raw,
                                  in_memory=not broadcast)
        return self.get_block_result(block,
                                     broadcast,
                                     source_seed=source_seed,
                                     source_index=source_index)

    def create_change_block(self,
                            source_seed,
                            source_index,
                            new_rep,
                            broadcast=True):
        block = self.create_block("change",
                                  source_seed=source_seed,
                                  source_index=source_index,
                                  link="0" * 64,
                                  representative=new_rep,
                                  in_memory=not broadcast)

        return self.get_block_result(block,
                                     broadcast,
                                     source_seed=source_seed,
                                     source_index=source_index)

    def create_change_block_pkey(self,
                                 source_private_key,
                                 new_rep,
                                 broadcast=True):
        block = self.create_block("change",
                                  source_private_key=source_private_key,
                                  link="0" * 64,
                                  representative=new_rep,
                                  in_memory=not broadcast)
        return self.get_block_result(block, broadcast)

    def create_send_block_pkey(self,
                               private_key,
                               destination_account,
                               amount_per_chunk_raw,
                               broadcast=True):

        block = self.create_block("send",
                                  source_private_key=private_key,
                                  destination_account=destination_account,
                                  amount_raw=amount_per_chunk_raw,
                                  in_memory=not broadcast)
        return self.get_block_result(block, broadcast)

    def create_epoch_block(self,
                           epoch_link,
                           genesis_private_key,
                           genesis_account,
                           broadcast=True):

        #account_info = self.account_info(genesis_account)
        #epoch_block = self.block_create(account_info["balance"],genesis_account, genesis_private_key, account_info["representative"],epoch_link,account_info["frontier"] )
        #print(epoch_block)
        block = self.create_block("epoch",
                                  source_private_key=genesis_private_key,
                                  destination_account=genesis_account,
                                  link=epoch_link,
                                  in_memory=not broadcast)

        return self.get_block_result(block, broadcast)



from nanomock.modules.nl_rpc import NanoRpc
from nanomock.modules.nl_nanolib import raw_high_precision_percent
from nanomock.modules.nl_parse_config import ConfigParser
from nanomock.internal.utils import get_mock_logger


class InitialBlocks:

    def __init__(self, config_parser: ConfigParser, rpc_url, logger=None):
        logger = logger or get_mock_logger()
        self.logger = logger
        self.api = NanoRpc(rpc_url)
        self.conf_p = config_parser

    def __epoch_link(self, epoch: int):
        message = f"epoch v{epoch} block"
        as_hex = bytearray(message, "ascii").hex()
        link = as_hex.upper().ljust(64, '0')
        return link

    def __publish_epochs(self):
        e = 1
        self.__log_active_difficulty()
        while e <= self.conf_p.get_all()["epoch_count"]:
            link = self.__epoch_link(e)
            epoch_block = self.api.create_epoch_block(
                link,
                self.conf_p.get_genesis_account_data()["private"],
                self.conf_p.get_genesis_account_data()["account"],
            )
            self.logger.append_log(
                "InitialBlocks", "INFO",
                "EPOCH {} sent by genesis : HASH {}".format(
                    e, epoch_block["hash"]))
            self.__log_active_difficulty()
            e += 1
        pass

    def __log_active_difficulty(self):
        diff = self.api.get_active_difficulty()
        self.logger.append_log(
            "InitialBlocks", "INFO",
            f'current_diff : [{diff["network_current"]}]  current_receive_diff: [{diff["network_receive_current"]}]'
        )

    def __publish_canary(self):
        fv_canary_send_block = self.api.create_send_block_pkey(
            self.conf_p.get_genesis_account_data()["private"],
            self.conf_p.get_canary_account_data()["account"], 1)
        self.logger.append_log(
            "InitialBlocks", "INFO",
            "SEND FINAL VOTES CANARY BLOCK FROM {} To {} : HASH {}".format(
                self.conf_p.get_genesis_account_data()["account"],
                self.conf_p.get_canary_account_data()["account"],
                fv_canary_send_block["hash"]))

        fv_canary_open_block = self.api.create_open_block(
            self.conf_p.get_canary_account_data()["account"],
            self.conf_p.get_canary_account_data()["private"], 1,
            self.conf_p.get_genesis_account_data()["account"],
            fv_canary_send_block["hash"])
        self.logger.append_log(
            "InitialBlocks", "INFO",
            "OPENED CANARY ACCOUNT {} : HASH {}".format(
                self.conf_p.get_canary_account_data()["account"],
                fv_canary_open_block["hash"]))

    def __send_to_burn(self):
        if "burn_amount" not in self.conf_p.get_all():
            self.logger.debug("[burn_amount] is not set. exit send_to_burn()")
            return False

        genesis_balance = int(
            self.api.check_balance(self.conf_p.get_genesis_account_data()
                                   ["account"])["balance_raw"])
        if int(self.conf_p.get_all()["burn_amount"]) > genesis_balance:
            self.logger.append_log(
                "InitialBlocks", "WARNING",
                "[burn_amount] exceeds genesis balance. exit send_to_burn()")
            return False

        send_block = self.api.create_send_block_pkey(
            self.conf_p.get_genesis_account_data()["private"],
            self.conf_p.get_burn_account_data()["account"],
            self.conf_p.get_all()["burn_amount"])

        self.logger.append_log(
            "InitialBlocks", "INFO",
            "SENT {:>40} FROM {} To {} : HASH {}".format(
                send_block["amount_raw"],
                self.conf_p.get_genesis_account_data()["account"],
                self.conf_p.get_burn_account_data()["account"],
                send_block["hash"]))

    def __convert_weight_percentage_to_balance(self):
        genesis_balance = int(
            self.api.check_balance(
                self.conf_p.get_genesis_account_data()["account"],
                include_only_confirmed=False)["balance_raw"])
        genesis_remaing = genesis_balance

        for node_conf in self.conf_p.get_nodes_config():

            if "vote_weight_percent" not in node_conf and "balance" not in node_conf:
                continue  #skip genesis that was added as node
            if "vote_weight_percent" in node_conf:
                node_conf["balance"] = raw_high_precision_percent(
                    genesis_balance, node_conf["vote_weight_percent"])
            node_conf["balance"] = int(node_conf["balance"])

            if genesis_remaing <= 0:
                self.logger.append_log(
                    "InitialBlocks", "WARNING",
                    f'No Genesis funds remaining! Account [{node_conf["account_data"]["account"]}] will not be opened!'
                )
                #self.conf_p["node_account_data"].remove(node_account_data)
                continue
            if genesis_remaing < node_conf["balance"]:
                self.logger.append_log(
                    "InitialBlocks", "WARNING",
                    f'Genesis remaining balance is too small! Send {genesis_remaing} instead of {node_conf["balance"]}.'
                )

            self.conf_p.set_node_balance(
                node_conf["name"], min(node_conf["balance"], genesis_remaing))
            genesis_remaing = max(0, genesis_remaing - node_conf["balance"])

    def __send_vote_weigh(self):

        for node_conf in self.conf_p.get_nodes_config():

            if "balance" not in node_conf:

                continue  #skip genesis that was added as node
            node_account_data = node_conf["account_data"]

            send_block = self.api.create_send_block_pkey(
                self.conf_p.get_genesis_account_data()["private"],
                node_account_data["account"], node_conf["balance"])

            self.logger.append_log(
                "InitialBlocks", "INFO",
                "SENT {:>40} FROM {} To {} : HASH {}".format(
                    send_block["amount_raw"],
                    self.conf_p.get_genesis_account_data()["account"],
                    node_account_data["account"], send_block["hash"]))

            open_block = self.api.create_open_block(
                node_account_data["account"], node_account_data["private"],
                node_conf["balance"], node_account_data["account"],
                send_block["hash"])

            self.logger.append_log(
                "InitialBlocks", "INFO",
                "OPENED PR ACCOUNT {} : HASH {}".format(
                    node_account_data["account"], open_block["hash"]))

    def create_node_wallet(self,
                           rpc_url,
                           node_name,
                           private_key=None,
                           seed=None):
        api = NanoRpc(rpc_url)

        if private_key != None:
            wallet = api.wallet_create(None)["wallet"]
            account = api.wallet_add(wallet, private_key)["account"]
        if seed != None:
            wallet = api.wallet_create(seed)["wallet"]
            account = api.get_account_data(seed, 0)["account"]
        self.logger.append_log(
            "InitialBlocks", "INFO",
            f"WALLET {wallet} CREATED FOR {node_name} WITH ACCOUNT {account}")

    def publish_initial_blocks(self):
        self.__publish_epochs()
        self.__publish_canary()
        self.__send_to_burn()
        self.__convert_weight_percentage_to_balance()
        self.__send_vote_weigh()
        return self.logger.pop("InitialBlocks")



import platform
from nanomock.internal.utils import get_mock_logger


class FeatureToggle:

    def __init__(self):
        self.os_name = platform.system()
        self.logger = get_mock_logger()
        #this way, if we remove the entry, the feature is enabled, even if we forget to remove the toggle code
        self.disabled_features = {
            "config_blkio": self.os_name == 'Darwin',
        }

    def is_feature_disabled(self, feature_name, custom_warning=None):
        disabled = self.disabled_features.get(feature_name, False)

        if not disabled:
            return False

        custom_warning = custom_warning or f"Feature {feature_name} is disabled for {self.os_name}"
        self.logger.warning(custom_warning)
        return True


toggle = FeatureToggle()


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="nanomock",
      version="0.0.11",
      author="gr0vity",
      description="Create local dockerized nano-currency networks",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/gr0vity-dev/nanomock",
      packages=find_packages(exclude=["unit_tests"]),
      include_package_data=True,
      install_requires=[
          "pyyaml",
          "tomli_w",
          "tomli",
          "oyaml",
          "nanolib",
          "extradict",
          "requests",
          'bitmath',
      ],
      entry_points={
          'console_scripts': [
              'nanomock=nanomock.main:main',
          ],
      })



import os
from pathlib import Path
import logging
from typing import List, Optional
from .internal.dependency_checker import DependencyChecker
from .modules.nl_parse_config import ConfigParser, ConfigReadWrite
from .internal.nl_initialise import InitialBlocks
from .docker import create_docker_interface
from .modules.nl_rpc import NanoRpc
from .internal.utils import log_on_success, get_mock_logger, shutil_rmtree, extract_packaged_services_to_disk, subprocess_run_capture_output
from typing import List, Dict, Optional, Tuple, Union
import concurrent.futures
from math import floor
import json
import time
import inspect

logger = None


def init_logger():
    global logger
    logger = logger or get_mock_logger()


class NanoLocalManager:

    def __init__(self, dir_path, project_name, config_file="nl_config.toml"):
        init_logger()
        self.command_mapping = self._initialize_command_mapping()
        self.conf_p = ConfigParser(dir_path,
                                   config_file=config_file,
                                   logger=logger)
        self.conf_rw = ConfigReadWrite()
        self.dependency_checker = DependencyChecker()
        self.dependency_checker.check_dependencies()

        self.dir_path = dir_path
        self.nano_nodes_path = self.conf_p.nano_nodes_path
        self.compose_yml_path = self.conf_p.compose_out_path
        self.compose_env_path = os.path.join(self.nano_nodes_path,
                                             "dc_nano_local_env")
        self.project_name = project_name
        self.services_dir = self.conf_p.services_dir
        self.nodes_data_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest"
        self.config_node_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest/config-node.toml"
        self.config_rpc_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest/config-rpc.toml"

        os.makedirs(self.nano_nodes_path, exist_ok=True)

        self.docker_interface = create_docker_interface(
            self.compose_yml_path, self.project_name)

    def _initialize_command_mapping(self):
        #(command_method , validation_method)
        return {
            'create': (self.create_docker_compose_file, None),
            'start': (self.start_containers, None),
            'start_nodes': (self.start_all_nodes, None),
            'status': (self.network_status, None),
            'restart': (self.restart_containers, None),
            'reset': (self.reset_nodes_data, None),
            'init': (self.init_nodes, None),
            'init_wallets': (self.init_wallets, None),
            'stop': (self.stop_containers, None),
            'stop_nodes': (self.stop_all_nodes, None),
            'remove': (self.remove_containers, None),
            'down': (self.remove_containers, None),
            'destroy': (lambda: self.destroy(remove_files=True), None),
            'rpc': (self.run_rpc, self._validator_rpc),
            'conf_edit': (self.conf_edit, self._validator_conf_edit)
        }

    def _get_default(self, config_name):
        """ Load config with default values"""
        #minimal node config if no file is provided in the nl_config.toml
        if config_name == "config_node":
            default_config_path = os.path.join(self.services_dir,
                                               "default_config-node.toml")
            return self.conf_rw.read_toml(default_config_path,
                                          is_packaged=True)
        elif config_name == "config_rpc":
            default_rpc_path = os.path.join(self.services_dir,
                                            "default_config-rpc.toml")
            return self.conf_rw.read_toml(default_rpc_path, is_packaged=True)
        else:
            return {}

    def _generate_docker_compose_env_file(self):
        env_variables = self.conf_p.get_docker_compose_env_variables()
        self.conf_rw.write_list(f'{self.compose_env_path}', env_variables)

    def _generate_docker_compose_yml_file(self):
        self.conf_p.set_docker_compose()
        self.conf_p.write_docker_compose()

    def _set_config_node_file(self, node_name):
        config_node = self.conf_p.get_config_from_path(node_name,
                                                       "config_node_path")
        if config_node is None:
            logger.warning(
                "No config-node.toml found. minimal version was created")
            config_node = self._get_default("config_node")

        config_node["node"][
            "preconfigured_peers"] = self.conf_p.preconfigured_peers
        config_node["node"]["enable_voting"] = self.conf_p.is_voting_enabled(
            node_name)
        return config_node

    def _generate_config_node_file(self, node_name):
        config_node = self._set_config_node_file(node_name)
        self.conf_rw.write_toml(
            self.config_node_path.format(node_name=node_name), config_node)

    def _generate_config_rpc_file(self, node_name):
        config_rpc = self.conf_p.get_config_from_path(node_name,
                                                      "config_rpc_path")
        if config_rpc is None:
            logger.warning(
                "No config-rpc.toml found. minimal version was created")
            config_rpc = self._get_default("config_rpc")

        self.conf_rw.write_toml(
            self.config_rpc_path.format(node_name=node_name), config_rpc)

    def _generate_nanomonitor_config_file(self, node_name):
        if self.conf_p.get_config_value("nanomonitor_enable"):
            self.conf_p.write_nanomonitor_config(node_name)

    def _create_node_folders(self, node_name):
        nano_node_path = os.path.join(self.nano_nodes_path, node_name)
        nano_test_path = os.path.join(nano_node_path, "NanoTest")

        os.makedirs(nano_node_path, exist_ok=True)
        os.makedirs(nano_test_path, exist_ok=True)

        if self.conf_p.get_config_value("nanomonitor_enable"):
            nano_monitor_path = os.path.join(nano_node_path, "nanoNodeMonitor")
            os.makedirs(nano_monitor_path, exist_ok=True)

    def _prepare_nodes(self):
        #prepare genesis
        for node_name in self.conf_p.get_nodes_name():
            self._prepare_node_env(node_name)

    def _prepare_node_env(self, node_name):
        node_name = node_name.lower(
        )  # docker-compose requires lower case names
        self._create_node_folders(node_name)
        self._generate_config_node_file(node_name)
        self._generate_config_rpc_file(node_name)
        self._generate_nanomonitor_config_file(node_name)

    def _online_containers(self, node_names: List[str]) -> List[str]:
        online_containers = []
        for container in node_names:
            cmd = f"docker ps |grep {container}$ | wc -l"
            res = subprocess_run_capture_output(cmd)
            online = int(res.stdout.strip())

            if online == 1:
                online_containers.append(container)

        return online_containers

    def _count_online_containers(self, node_names: List[str]) -> int:
        online_containers = self._online_containers(node_names)
        online_count = len(online_containers)
        return online_count

    def _get_nodes_block_counts(
            self, nodes_name: List[str]) -> List[Dict[str, Union[str, int]]]:
        nodes_block_count = []

        nodes_rpc = ([
            NanoRpc(self.conf_p.get_node_rpc(node)) for node in nodes_name
        ] if nodes_name is not None else self._get_all_rpc())

        for nano_rpc in nodes_rpc:
            block_count = nano_rpc.block_count()
            version_rpc_call = nano_rpc.version()
            if not (block_count or version_rpc_call):
                continue
            node_name = self.conf_p.get_node_name_from_rpc_url(nano_rpc)
            count = int(block_count["count"])
            nodes_block_count.append({
                "node_name": node_name,
                "count": count,
                "cemented": block_count["cemented"],
                "version": version_rpc_call
            })

        return nodes_block_count

    def _generate_network_status_report(
            self, nodes: List[str],
            nodes_block_count: List[Dict[str, Union[str, int]]]) -> str:

        if not nodes_block_count:
            return ""

        max_count = max(int(bc["count"]) for bc in nodes_block_count)

        def get_node_data(
                node_name: str) -> Union[Dict[str, Union[str, int]], None]:
            for bc in nodes_block_count:
                if bc["node_name"] == node_name:
                    return bc
            return None

        def format_report_line(bc: Dict[str, Union[str, int]]) -> str:
            node_version = f'{bc["version"]["node_vendor"]} {bc["version"]["build_info"].split(" ")[0]}'
            synced_percentage = floor(
                int(bc["cemented"]) / max_count * 10000) / 100
            return '{:<16} {:<20} {:>6.2f}% synced | {}/{} blocks cemented'.format(
                bc["node_name"], node_version, synced_percentage,
                bc["cemented"], bc["count"])

        report = []
        for node_name in nodes:
            node_data = get_node_data(node_name)
            if node_data:
                report_line = format_report_line(node_data)
            else:
                report_line = f'{node_name} [down]'
            report.append(report_line)

        return '\n' + '\n'.join(report)

    def _get_all_rpc(self):
        return [NanoRpc(x) for x in self.conf_p.get_nodes_rpc()]

    def _is_rpc_available(self,
                          container: str,
                          timeout: int = 3) -> Tuple[str, bool]:
        rpc_url = self.conf_p.get_node_rpc(container)
        try:
            nano_rpc = NanoRpc(rpc_url)
            logging.info("call _is_rpc_available")
            if nano_rpc.block_count(max_retry=0):
                return container, True
        except Exception as e:
            logging.warning(
                f"RPC {rpc_url} not yet reachable for node {container}: {str(e)}"
            )

        return container, False

    def _wait_for_rpc_availability(self,
                                   nodes_name: List[str] = None,
                                   wait: bool = True,
                                   timeout: int = 10,
                                   max_timeout_s: int = 15) -> None:
        start_time = time.time()

        nodes_name = nodes_name or self.conf_p.get_nodes_name()
        nodes_to_check = nodes_name.copy()

        def get_unavailable_nodes(nodes: List[str], timeout: int) -> List[str]:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                results = list(
                    executor.map(self._is_rpc_available, nodes,
                                 [timeout] * len(nodes)))
            return [
                container for container, available in results if not available
            ]

        while len(nodes_to_check) > 0:
            if not wait:
                break

            nodes_to_check = get_unavailable_nodes(nodes_to_check, timeout)

            if time.time() - start_time > max_timeout_s:
                raise ValueError(
                    f"TIMEOUT: RPCs not reachable for nodes {nodes_to_check}")

            if len(nodes_to_check) > 0:
                time.sleep(1)

        logger.info(f"Nodes {nodes_name} reachable")

    def online_containers_status(self, online_containers: List[str],
                                 total_nodes) -> str:

        online_count = len(online_containers)
        if online_count == total_nodes:
            return f"All {online_count} containers online"
        else:
            return f"{online_count}/{total_nodes} containers online"

    def conf_edit(self, payload):
        self.conf_p.modify_nanolocal_config(payload["path"], payload["value"])
        return True

    @log_on_success
    def network_status(self, nodes_name: Optional[List[str]] = None) -> str:
        if nodes_name == []:
            return ""

        nodes_name = self.conf_p.get_nodes_name()
        online_containers = self._online_containers(nodes_name)
        status_msg = self.online_containers_status(online_containers,
                                                   len(nodes_name))
        # if "All" not in status_msg:
        #     return status_msg

        nodes_block_count = self._get_nodes_block_counts(online_containers)
        return status_msg + self._generate_network_status_report(
            nodes_name, nodes_block_count)

    @log_on_success
    def create_docker_compose_file(self):
        extract_packaged_services_to_disk(self.nano_nodes_path)
        self._prepare_nodes()
        self._generate_docker_compose_env_file()
        self._generate_docker_compose_yml_file()
        self.docker_interface.create_network(self.conf_p.get_network_name())
        self._initilaise_services_configs(self.conf_p)

        logger.success(
            f"Docker Compose file created at {self.compose_yml_path}")
        return None, "\n".join(self.conf_p.get_enabled_services())

    def _initilaise_services_configs(self, config: ConfigParser):
        if bool(config.get_config_value("nanocap_enable")):
            network_name = config.get_network_name()
            device_ip = self.docker_interface.get_network_gateway(network_name)
            config.set_nanocap_device_ip(device_ip)

    def _validator_rpc(self, nodes=None, payload=None):
        if not payload:
            raise ValueError(
                "The --payload argument is required for the 'rpc' command.")
        return nodes, payload

    def _validator_conf_edit(self, nodes=None, payload=None):
        if payload is None:
            raise ValueError(
                "payload must be provided '{\"path\" : ... , \"value\": ...}'")
        if "path" not in payload:
            raise ValueError("key \"path\" must be provided")
        if "value" not in payload:
            raise ValueError("key \"value\" must be provided")

        return nodes, payload

    @log_on_success
    def run_rpc(self, payload=None, nodes=None):
        responses = []
        if nodes is None:
            nodes = self.conf_p.get_nodes_name()

        for node in nodes:
            node_rpc = NanoRpc(self.conf_p.get_node_rpc(node))
            response = node_rpc.post_with_auth(payload)
            responses.append(response)

        return None, json.dumps(responses, indent=2)

    @log_on_success
    def init_wallets(self):
        #self.start_nodes('all')  #fixes a bug on mac m1
        init_blocks = InitialBlocks(self.conf_p,
                                    self.conf_p.get_nodes_rpc()[0])
        for node_name in self.conf_p.get_nodes_name():
            if node_name == self.conf_p.get_genesis_node_name():
                init_blocks.create_node_wallet(
                    self.conf_p.get_node_config(
                        self.conf_p.get_genesis_node_name())["rpc_url"],
                    self.conf_p.get_genesis_node_name(),
                    private_key=self.conf_p.config_dict["genesis_key"])
            else:
                init_blocks.create_node_wallet(
                    self.conf_p.get_node_config(node_name)["rpc_url"],
                    node_name,
                    seed=self.conf_p.get_node_config(node_name)["seed"])

        #return without logging for unit tests
        return init_blocks.logger.pop("InitialBlocks")

    @log_on_success
    def init_nodes(self):
        self.init_wallets()
        init_blocks = InitialBlocks(self.conf_p,
                                    self.conf_p.get_nodes_rpc()[0])
        return init_blocks.publish_initial_blocks()

    @log_on_success
    def reset_nodes_data(self, nodes: Optional[List[str]] = None):
        self.stop_containers(nodes)
        nodes_to_process = nodes or ['.']

        for node in nodes_to_process:
            node_path = Path(self.nano_nodes_path) / node if nodes else Path(
                self.nano_nodes_path)
            files_to_remove = ['data.ldb', 'wallets.ldb']

            for file_pattern in files_to_remove:
                for file_path in node_path.rglob(file_pattern):
                    file_path.unlink()

    def init_containers(self):
        self.create_docker_compose_file()
        self.build_containers()

    @log_on_success
    def restart_containers(self, nodes: Optional[List[str]] = None):
        return None, self.docker_interface.compose_restart(nodes)

    @log_on_success
    def build_containers(self):
        return None, self.docker_interface.compose_build()

    @log_on_success
    def start_containers(self, nodes: Optional[List[str]] = None):
        self.docker_interface.compose_start(nodes)
        self._wait_for_rpc_availability(nodes)

    @log_on_success
    def start_all_nodes(self):
        nodes = self.conf_p.get_nodes_name()
        self.docker_interface.compose_start(nodes)
        self._wait_for_rpc_availability(nodes)

    @log_on_success
    def stop_containers(self, nodes: Optional[List[str]] = None):
        #by default, stops all containers (also services like nanolooker, monitor or prom-exporter)
        self.docker_interface.compose_stop(nodes)

    @log_on_success
    def stop_all_nodes(self):
        #stops all nodes but leaves services running
        nodes = self.conf_p.get_nodes_name()
        self.docker_interface.compose_stop(nodes)

    @log_on_success
    def remove_containers(self):
        containers = self.conf_p.get_conatiners_name()
        self.docker_interface.compose_down()
        started_count = self.docker_interface.container_count(containers)
        removed_count = len(containers) - started_count
        return f"{removed_count} containers have been removed"

    @log_on_success
    def destroy(self, remove_files=False):
        # Stop and remove containers
        self.remove_containers()

        # Remove the created files and folders if remove_files is True
        if remove_files:
            return None, shutil_rmtree(self.nano_nodes_path)

    @log_on_success
    def update(self):
        self.docker_interface.compose_pull()
        # Remove containers and networks
        self.remove_containers()
        self.build_containers()

    def _filter_args(self, func, **kwargs):
        sig = inspect.signature(func)
        filtered_args = {
            k: v
            for k, v in kwargs.items() if k in sig.parameters
        }
        return filtered_args

    def execute_command(self, command, nodes=None, payload=None):
        if command not in self.command_mapping:
            raise ValueError(f"Invalid command: {command}")

        command_func, validator_func = self.command_mapping[command]

        if validator_func is not None:
            filtered_validator_args = self._filter_args(validator_func,
                                                        nodes=nodes,
                                                        payload=payload)
            validated_nodes, validated_payload = validator_func(
                **filtered_validator_args)
        else:
            validated_nodes, validated_payload = nodes, payload

        filtered_command_args = self._filter_args(command_func,
                                                  nodes=validated_nodes,
                                                  payload=validated_payload)
        command_func(**filtered_command_args)


if __name__ == "__main__":
    pass



import subprocess
import functools
import logging
from typing import Callable, List, Any, Tuple
from importlib.metadata import version, PackageNotFoundError
from importlib import resources
from pathlib import Path
from typing import Tuple
import tomli
import oyaml as yaml
import json
import bitmath
import shutil


def get_mock_logger():
    logger.setLevel(logging.INFO)
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
        logger.addHandler(handler)
    return logger


class NanoMockLogger(logging.Logger):
    LOG_STORE = {}
    SUCCESS_LEVEL_NUM = 25
    logging.addLevelName(SUCCESS_LEVEL_NUM, "SUCCESS")

    def success(self, message, *args, **kws):
        self.log(self.SUCCESS_LEVEL_NUM, message, *args, **kws)

    def dynamic(self, log_level, message):
        log_level = getattr(logger, log_level, logging.INFO)
        logger.log(log_level, message)

    def append_log(self, log_key, log_level, message):
        self.LOG_STORE.setdefault(log_key, [])
        self.LOG_STORE[log_key].append(message)
        self.dynamic(log_level, message)

    def pop(self, log_key):
        return self.LOG_STORE.pop(log_key)


# Set the custom logger as the root logger
logging.setLoggerClass(NanoMockLogger)
logger = logging.getLogger(__name__)


def log_on_success(func: Callable) -> Callable:

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        result = func(*args, **kwargs)

        if isinstance(result, tuple) and len(result) == 2:
            value, log_message = result
            logger.success(log_message)
            return value
        elif result is None:
            logger.success(func.__name__)
            return None
        else:
            logger.success(result)
            return result

    return wrapper


def is_packaged_version():
    is_packaged = None
    try:
        _ = version('nanomock')
        is_packaged = True
    except PackageNotFoundError:
        is_packaged = False

    return is_packaged


def _convert_to_dotted_str(path: str) -> str:
    return path.replace("/", ".")


def _convert_to_dotted_path(path: Path, base_dir: str,
                            replacement: str) -> Tuple[str, str]:
    parts = path.parts
    new_parts = [replacement if p == base_dir else p for p in parts[:-1]]
    dot_separated_file_path = '.'.join(new_parts)
    file_name = parts[-1]
    return dot_separated_file_path, file_name


def _split_file_from_path(path: str) -> Tuple[str, str]:
    parts = path.split(".")
    file_name = '.'.join(parts[-2::])
    dot_separated_file_path = '.'.join(parts[0:-2])

    return dot_separated_file_path, file_name


def _read_data(dotted_path: str, file_name: str):

    with resources.open_text(dotted_path, file_name) as file:
        data = file.read()
    return data


def read_from_package_if_needed(read_method):

    @functools.wraps(read_method)
    def wrapper(self, path, *args, is_packaged=False, **kwargs):
        if is_packaged_version() and is_packaged:
            # Convert path to dotted path and read data using read_data method
            dotted_path = _convert_to_dotted_str(path)
            file_path, file_name = _split_file_from_path(dotted_path)
            data = _read_data(file_path, file_name)

            # If the read method is expected to return a different data format,
            # you can add the conversion logic here based on the read_method.__name__
            if read_method.__name__ == "read_json":
                return json.loads(data)
            elif read_method.__name__ == "read_toml":
                return tomli.loads(data)
            elif read_method.__name__ == "read_yaml":
                return yaml.safe_load(data)
            else:
                # For read_file, the data is already in the correct format (list of lines)
                return data.splitlines()
        else:
            # If not a packaged version, call the original read_method
            return read_method(self, path, *args, is_packaged=False, **kwargs)

    return wrapper


def find_device_for_path(path: str) -> str:
    path = Path(path).resolve()
    df_output = subprocess.check_output(['df', path],
                                        universal_newlines=True).splitlines()

    # The first line is the header, and the second line contains the information we need
    device = df_output[1].split()[0]

    return device


def convert_to_bytes(size_string) -> int:
    if str(size_string).isnumeric():
        return int(size_string)

    size = bitmath.parse_string(size_string)
    return int(size.to_Byte())


def extract_packaged_services_to_disk(destination_path: Path):
    with resources.path("nanomock", "__init__.py") as src_dir:
        src_dir = src_dir.parent / "internal" / "data" / "services"
        dest_dir = destination_path / "services"

        if src_dir.exists():
            dest_dir.mkdir(parents=True, exist_ok=True)
            shutil.copytree(src_dir, dest_dir, dirs_exist_ok=True)
            logger.info(
                "nanomock data has been copied to your current working directory."
            )
        else:
            logger.error("Error: nanomock data not found.")


def shutil_rmtree(path: Path):
    shutil.rmtree(path)
    return f"Removed directory: {path}"


def subprocess_run_capture_output(cmd, shell=True, cwd=None):

    result = subprocess.run(cmd,
                            shell=shell,
                            check=True,
                            capture_output=True,
                            text=True,
                            cwd=cwd)
    return result


from .interface import DockerInterface
from .mixin import DockerMixin
import subprocess


class MacDockerInterface(DockerMixin,DockerInterface):

    @classmethod
    def get_docker_gateway_ip(cls) -> str:
        command = "ifconfig | grep -A1 'docker' | grep inet | awk '{print $2}'"
        output = subprocess.check_output(command, shell=True, text=True)
        return output.strip()
